package ch.unige.aou.service;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;
import javax.mail.MessagingException;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import ch.unige.solidify.exception.SolidifyRuntimeException;
import ch.unige.solidify.service.EmailService;
import ch.unige.solidify.service.MessageService;

import ch.unige.aou.business.PublicationService;
import ch.unige.aou.business.ResearchGroupService;
import ch.unige.aou.config.AouProperties;
import ch.unige.aou.message.EmailMessage;
import ch.unige.aou.model.publication.Publication;
import ch.unige.aou.model.settings.ResearchGroup;

@Service
public class EmailProcessingService extends AouService {

  private static final Logger log = LoggerFactory.getLogger(EmailProcessingService.class);

  private final EmailService emailService;
  private final PublicationService publicationService;
  private final ResearchGroupService researchGroupService;

  private final String submissionPortalHomepage;
  private final String homepage;
  private final int delaySinceLastCheck;

  public EmailProcessingService(MessageService messageService, AouProperties aouProperties, EmailService emailService,
          PublicationService publicationService, ResearchGroupService researchGroupService) {
    super(messageService);
    this.emailService = emailService;
    this.publicationService = publicationService;
    this.researchGroupService = researchGroupService;
    this.submissionPortalHomepage = aouProperties.getParameters().getSubmissionPortalHomepage();
    this.homepage = aouProperties.getParameters().getHomepage();
    this.delaySinceLastCheck = aouProperties.getDelaySinceLastCheck();
  }

  @Transactional
  public void processMessage(EmailMessage emailMessage) {
    Map<String, Object> parameterList = new HashMap<>();
    try {
      switch (emailMessage.getTemplate()) {
        case NEW_DEPOSITS_TO_VALIDATE:
          this.sendEmailWithNewDepositsToValidate(emailMessage);
          break;
        case MY_DEPOSIT_REQUIRES_FEEDBACK:
          this.sendEmailToDepositRequiringFeedbackDepositor(emailMessage);
          break;
        case MY_INDIRECT_PUBLICATION_VALIDATED:
          this.sendEmailToValidatedDepositsContributor(emailMessage);
          break;
        case MY_PUBLICATIONS_VALIDATED:
          this.sendEmailToValidatedDepositsDepositor(emailMessage);
          break;
        case MY_PUBLICATIONS_REJECTED:
          this.sendEmailToRejectedDepositsDepositor(emailMessage);
          break;
        case NEW_COMMENT_FOR_VALIDATORS_DEPOSIT:
          this.sendEmailToValidatorsOnNewCommentInDeposit(emailMessage);
          break;
        case NEW_SUBSCRIPTION:
        case REMOVE_SUBSCRIPTION:
          parameterList.put("notificationTypes", emailMessage.getParameters());
          this.emailService.sendEmailWithTemplate(emailMessage.getTemplate().toString().toLowerCase(),
                  Collections.singletonList(emailMessage.getTo()), emailMessage.getTemplate().getSubject(), parameterList);
          break;
        case FORGOTTEN_PUBLICATIONS_TO_VALIDATE:
          this.sendEmailWithPublicationsList(emailMessage, this.delaySinceLastCheck);
          break;
        case SYNCHRONIZED_RESEARCH_GROUPS:
          this.sendEmailWithSynchronizedResearchGroups(emailMessage);
          break;
        default:
          throw new SolidifyRuntimeException("Email template " + emailMessage.getTemplate() + " not supported");
      }

    } catch (NoSuchElementException | MessagingException e) {
      log.error("Cannot send email ({}): {}", emailMessage.getTemplate().getSubject(), e.getMessage(), e);
    }
  }

  private void sendEmailWithPublicationsList(EmailMessage emailMessage) throws MessagingException {
    List<Publication> publications = new ArrayList<>();
    for (String publicationId : emailMessage.getParameters()) {
      publications.add(this.publicationService.findOne(publicationId));
    }

    Map<String, Object> parameters = this.getEmailDefaultParameters();
    parameters.put("publications", publications);

    this.emailService.sendEmailWithTemplate(emailMessage.getTemplate().toString().toLowerCase(), Collections.singletonList(emailMessage.getTo()),
            emailMessage.getTemplate().getSubject(), parameters);
  }

  private void sendEmailWithPublicationsList(EmailMessage emailMessage, int delay) throws MessagingException {
    List<Publication> publications = new ArrayList<>();
    for (String publicationId : emailMessage.getParameters()) {
      publications.add(this.publicationService.findOne(publicationId));
    }

    Map<String, Object> parameters = this.getEmailDefaultParameters();
    parameters.put("publications", publications);
    parameters.put("delay", delay);
    this.emailService.sendEmailWithTemplate(emailMessage.getTemplate().toString().toLowerCase(), Collections.singletonList(emailMessage.getTo()),
            emailMessage.getTemplate().getSubject(), parameters);
  }

  private void sendEmailWithOnePublication(EmailMessage emailMessage) throws MessagingException {
    Publication publication = this.publicationService.findOne(emailMessage.getParameters()[0]);
    Map<String, Object> parameters = this.getEmailDefaultParameters();
    parameters.put("publication", publication);

    this.emailService.sendEmailWithTemplate(emailMessage.getTemplate().toString().toLowerCase(), Collections.singletonList(emailMessage.getTo()),
            emailMessage.getTemplate().getSubject(), parameters);
  }

  private void sendEmailWithNewDepositsToValidate(EmailMessage emailMessage) throws MessagingException {
    this.sendEmailWithPublicationsList(emailMessage);
  }

  private void sendEmailToDepositRequiringFeedbackDepositor(EmailMessage emailMessage) throws MessagingException {
    this.sendEmailWithOnePublication(emailMessage);
  }

  private void sendEmailToValidatedDepositsContributor(EmailMessage emailMessage) throws MessagingException {
    this.sendEmailWithPublicationsList(emailMessage);
  }

  private void sendEmailToValidatedDepositsDepositor(EmailMessage emailMessage) throws MessagingException {
    this.sendEmailWithPublicationsList(emailMessage);
  }

  private void sendEmailToRejectedDepositsDepositor(EmailMessage emailMessage) throws MessagingException {
    this.sendEmailWithPublicationsList(emailMessage);
  }

  private void sendEmailToValidatorsOnNewCommentInDeposit(EmailMessage emailMessage) throws MessagingException {
    this.sendEmailWithPublicationsList(emailMessage);
  }

  private void sendEmailWithSynchronizedResearchGroups(EmailMessage emailMessage) throws MessagingException {
    List<ResearchGroup> researchGroups = new ArrayList<>();
    for (String researchGroupId : emailMessage.getParameters()) {
      researchGroups.add(this.researchGroupService.findOne(researchGroupId));
    }

    Map<String, Object> parameters = this.getEmailDefaultParameters();
    parameters.put("researchGroups", researchGroups);

    this.emailService.sendEmailWithTemplate(emailMessage.getTemplate().toString().toLowerCase(), Collections.singletonList(emailMessage.getTo()),
            emailMessage.getTemplate().getSubject(), parameters);
  }

  private Map<String, Object> getEmailDefaultParameters() {
    Map<String, Object> parameters = new HashMap<>();
    parameters.put("submissionPortalHomepage", this.submissionPortalHomepage);
    parameters.put("homepage", this.homepage);
    return parameters;
  }
}
