package ch.unige.aou.service;

import java.io.IOException;

import org.springframework.boot.logging.LogLevel;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import ch.unige.solidify.exception.SolidifyCheckingException;
import ch.unige.solidify.exception.SolidifyProcessingException;
import ch.unige.solidify.service.MessageService;

import ch.unige.aou.business.PublicationService;
import ch.unige.aou.model.publication.Publication;
import ch.unige.aou.model.publication.Publication.PublicationStatus;
import ch.unige.aou.service.storage.StorageService;

@Service
public class PublicationProcessingService extends AouService {

  private final PublicationService publicationService;
  private final StorageService storageService;

  public PublicationProcessingService(MessageService messageService, PublicationService publicationService, StorageService storageService) {
    super(messageService);
    this.publicationService = publicationService;
    this.storageService = storageService;
  }

  /**
   * Return true if a new DocumentFileMessage must be sent to continue the DocumentFile processing
   *
   * @param documentFileMessage document file message object
   * @return boolean
   */
  @Transactional
  public Publication processPublication(String publicationId) {

    final Publication publication = this.publicationService.findOne(publicationId);

    this.logPublicationMessage(LogLevel.INFO, publication, "will be processed");

    try {
      switch (publication.getStatus()) {
        case SUBMITTED:
          this.processSubmittedPublication(publication);
          break;
        case CHECKED:
          this.processCheckedPublication(publication);
          break;
        case IN_PREPARATION:
          this.processInPreparationPublication(publication);
          break;
        default:
          break;
      }
    } catch (final SolidifyCheckingException e) {
      if (e.getCause() != null) {
        this.inError(publication, e.getCause().getMessage());
      } else {
        this.inError(publication, e.getMessage());
      }
    } catch (final SolidifyProcessingException e) {
      this.inError(publication, e.getMessage());
    } catch (final Exception e) {
      this.logPublicationMessage(LogLevel.ERROR, publication, "Processing error", e);
      this.inError(publication, e.getMessage());
    }

    return this.publicationService.save(publication);
  }

  private void processSubmittedPublication(Publication publication) throws IOException {
    this.storageService.checkMetadata(publication);
    publication.setStatus(PublicationStatus.CHECKED);
  }

  private void processCheckedPublication(Publication publication) {
    this.storageService.generateExportMetadata(publication);
    publication.setStatus(PublicationStatus.IN_PREPARATION);
  }

  private void processInPreparationPublication(Publication publication) throws IOException {
    this.storageService.storePublication(publication);
    publication.setStatus(PublicationStatus.COMPLETED);
  }

  private void logPublicationMessage(LogLevel logLevel, Publication publication, String message, Exception... e) {
    this.logObjectWithStatusMessage(logLevel, publication.getClass().getSimpleName(), publication.getResId(),
            publication.getStatus().toString(), publication.getTitle(), message, e);
  }

  private void inError(Publication publication, String message) {
    publication.setErrorStatusWithMessage(message);
    this.logPublicationMessage(LogLevel.ERROR, publication, message);
  }
}
