package ch.unige.aou.test.admin;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.junit.jupiter.api.Assertions.assertThrows;

import org.junit.jupiter.api.Test;
import org.springframework.http.HttpStatus;
import org.springframework.web.client.HttpClientErrorException;

import ch.unige.aou.AouConstants;
import ch.unige.aou.model.publication.Comment;
import ch.unige.aou.model.publication.Publication;

class PublicationCommentIT extends AbstractAdminIT {

  @Override
  protected void createFixtures() {
    this.restClientTool.sudoRoot();
    this.createTopStructureFixture();
    this.createPublicationTypeAndSubtypeFixtures(false);
    this.createPublicationFixture();
    this.createCommentInPublicationTestFixture();
    this.restClientTool.exitSudo();
  }

  @Test
  void createInOtherPublicationTest() {
    //get publication
    Publication publication = this.getTestPublication();
    // Create a new Comment
    Comment comment = new Comment();
    comment.setText(this.getTemporaryTestLabel("comment test"));
    comment.setPublication(publication);
    comment.setOnlyForValidators(false);
    String publicationId = publication.getResId();
    HttpClientErrorException e = assertThrows(HttpClientErrorException.class, () -> {
      this.commentClientService.create(publicationId, comment);
    });
    assertEquals(HttpStatus.FORBIDDEN, e.getStatusCode());
  }

  @Test
  void findAllInOtherPublicationTest() {
    //get publication
    Publication publication = this.getTestPublication();

    // check that it is not possible to find comments in a publication that has not been created by the user(or being a validator)
    String publicationId = publication.getResId();
    HttpClientErrorException e = assertThrows(HttpClientErrorException.class, () -> {
      this.commentClientService.findAll(publicationId);
    });
    assertEquals(HttpStatus.FORBIDDEN, e.getStatusCode());
  }

  @Test
  void getOtherCommentTest() {
    // create publication and comments
    Publication publication = this.getTestPublication();
    this.restClientTool.sudoRoot();
    // Create a new Comment
    Comment comment = new Comment();
    comment.setText(this.getTemporaryTestLabel("comment test"));
    comment.setPublication(publication);
    comment.setOnlyForValidators(false);
    this.commentClientService.create(publication.getResId(), comment);
    this.restClientTool.exitSudo();

    //Get comments from this publication
    String publicationId = publication.getResId();
    String commentId = comment.getResId();
    HttpClientErrorException e = assertThrows(HttpClientErrorException.class, () -> {
      this.commentClientService.findOne(publicationId, commentId);
    });
    assertEquals(HttpStatus.FORBIDDEN, e.getStatusCode());

  }

  @Test
  void createTest() {
    //get publication
    Publication publication = this.createPublicationFixture(this.getTemporaryTestLabel("Publication to add comment"));
    // Create a new Comment
    Comment comment = new Comment();
    comment.setText(this.getTemporaryTestLabel("comment test"));
    comment.setPublication(publication);
    comment.setOnlyForValidators(false);
    comment = this.commentClientService.create(publication.getResId(), comment);
    // Test creation
    assertNotNull(comment.getResId());
  }

  @Test
  void createForValidatorTest() {
    //get publication
    Publication publication = this.createPublicationFixture(this.getTemporaryTestLabel("Publication to add comment"));
    // Create a new Comment
    Comment comment = new Comment();
    comment.setText(this.getTemporaryTestLabel("comment test"));
    comment.setPublication(publication);
    comment.setOnlyForValidators(true);
    String publicationId = publication.getResId();
    HttpClientErrorException e = assertThrows(HttpClientErrorException.class, () -> {
      this.commentClientService.create(publicationId, comment);
    });
    assertEquals(HttpStatus.BAD_REQUEST, e.getStatusCode());
  }

  @Test
  void findAllTest() {
    //get publication
    Publication publication = this.createPublicationFixture(this.getTemporaryTestLabel("Publication to find comment"));
    // Create a new Comment
    Comment comment = new Comment();
    comment.setText(this.getTemporaryTestLabel("comment test"));
    comment.setPublication(publication);
    comment.setOnlyForValidators(false);
    comment = this.commentClientService.create(publication.getResId(), comment);

    // findAll
    int sizeOfComments = this.commentClientService.findAll(publication.getResId()).size();

    // Test at least
    assertEquals(1, sizeOfComments);
  }

  @Test
  void deleteTest() {
    //get publication
    Publication publication = this.createPublicationFixture(this.getTemporaryTestLabel("Publication to delete comment"));
    // Create a new Comment
    Comment comment = new Comment();
    comment.setText(this.getTemporaryTestLabel("comment test"));
    comment.setPublication(publication);
    comment.setOnlyForValidators(false);

    comment = this.commentClientService.create(publication.getResId(), comment);

    // findAll
    int sizeOfCommentsBefore = this.commentClientService.findAll(publication.getResId()).size();

    // Test at least
    assertEquals(1, sizeOfCommentsBefore);

    // Delete
    this.commentClientService.delete(publication.getResId(), comment.getResId());
    int sizeOfCommentsAfter = this.commentClientService.findAll(publication.getResId()).size();

    // Test deletion
    assertEquals(0, sizeOfCommentsAfter);
  }

  private void clearCommentFixtures(String publicationId) {
    this.commentClientService.findAll(publicationId).stream()
            .filter(comment -> comment.getText().startsWith(AouConstants.TEMPORARY_TEST_DATA_LABEL))
            .forEach(comment -> this.commentClientService.delete(publicationId, comment.getResId()));
  }

  @Override
  protected void deleteFixtures() {
    this.restClientTool.sudoRoot();
    this.clearCommentFixtures();
    this.clearPublicationFixtures();
    this.clearPublicationSubtypeFixtures();
    this.clearPublicationTypeFixtures();
    this.clearStructureFixtures();
    this.restClientTool.exitSudo();
  }

}
