package ch.unige.aou.model.tool;

import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class CleanTool {

  /**
   * List of different existing hyphens
   * <p>
   * –   en dash                   0x2013  &ndash;
   * —   em dash                   0x2014  &mdash;
   * ‒   figure dash               0x2012  &#8210;
   * ﹘  small em dash             0xFE58  &#65112;
   * -   hyphen-minus              0x2D    &#45;
   * ֊   armenian hyphen           0x58A   &#1418;
   * ‐   hyphen                    0x2010  &#8208;
   * ‑   non-breaking hyphen       0x2011  &#8209;
   * ⁃   hyphen bullet             0x2043  &#8259;
   * ﹣  small hyphen-minus        0xFE63  &#65123;
   * －  fullwidth hyphen-minus    0xFF0D  &#65293;
   */
  public static final List<String> hyphensList = List.of("–", "—", "‒", "﹘", "-", "֊", "‐", "‑", "⁃", "﹣", "－");

  public static String cleanTitle(String value) {
    value = CleanTool.cleanString(value);
    value = CleanTool.removeFinalDot(value);
    return value;
  }

  public static String cleanString(String value) {
    value = CleanTool.trim(value);
    value = CleanTool.removeDoubleSpaces(value);
    return value;
  }

  public static String cleanPaging(String value) {
    value = CleanTool.trim(value);
    value = CleanTool.removeDoubleSpaces(value);
    value = CleanTool.removeBlankAroundHyphen(value);
    value = CleanTool.completeEndPageValue(value);
    return value;
  }

  /*******************************************************************************************/

  public static String trim(String value) {
    if (value != null) {
      value = value.trim();
    }
    return value;
  }

  public static String removeDoubleSpaces(String value) {
    if (value != null) {
      value = value.replaceAll("\\s+", " ");
    }
    return value;
  }

  public static String removeFinalDot(String value) {
    if (value != null && value.endsWith(".")) {
      value = value.substring(0, value.length() - 1);
    }
    return value;
  }

  /**
   * Replace hyphens of different kinds and remove surrounding spaces
   *
   * @param value
   * @return
   */
  public static String removeBlankAroundHyphen(String value) {
    if (value != null) {
      for (String hyphen : CleanTool.hyphensList) {
        if (hyphen.equals("-")) {
          // escape simple hyphen to be used in regex
          hyphen = "\\\\" + hyphen;
        }
        value = value.replaceAll("\\s*[" + hyphen + "]\\s*", "-");
      }
    }
    return value;
  }

  /**
   * Complete end page after hyphen such as :
   * <p>
   * "p. 123-45" --> "p. 123-145"
   * but
   * "p. 123-17" isn't updated as the result would make no sense.
   *
   * @param value
   * @return
   */
  public static String completeEndPageValue(String value) {
    if (value != null) {
      Pattern p = Pattern.compile("\\d+\\-\\d+");
      Matcher m = p.matcher(value);
      if (m.find()) {
        String pages = m.group(0);
        String startPage = pages.substring(0, pages.indexOf("-"));
        StringBuilder endPageBuilder = new StringBuilder(pages.substring(pages.indexOf("-") + 1));
        while (startPage.length() > endPageBuilder.length()) {
          int unitToComplete = startPage.length() - endPageBuilder.length() - 1;
          endPageBuilder.insert(0, startPage.charAt(unitToComplete));
        }
        if (Integer.valueOf(startPage) <= Integer.valueOf(endPageBuilder.toString())) {
          String newPages = startPage + "-" + endPageBuilder;
          value = value.replace(pages, newPages);
        }
      }
    }
    return value;
  }
}
