package ch.unige.aou.business;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.time.LocalDate;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.stream.Collectors;
import javax.transaction.Transactional;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.http.MediaType;
import org.springframework.stereotype.Service;
import org.springframework.util.StringUtils;
import org.springframework.validation.BindingResult;
import org.springframework.validation.FieldError;
import org.springframework.web.multipart.MultipartFile;

import ch.unige.solidify.exception.SolidifyFileDeleteException;
import ch.unige.solidify.exception.SolidifyRuntimeException;
import ch.unige.solidify.specification.SolidifySpecification;
import ch.unige.solidify.util.FileTool;
import ch.unige.solidify.util.StringTool;

import ch.unige.aou.AouConstants;
import ch.unige.aou.config.AouProperties;
import ch.unige.aou.model.Sortable;
import ch.unige.aou.model.publication.DocumentFile;
import ch.unige.aou.model.publication.DocumentFile.DocumentFileStatus;
import ch.unige.aou.model.publication.DocumentFileType;
import ch.unige.aou.model.publication.Publication;
import ch.unige.aou.model.rest.DocumentFileStatisticsDTO;
import ch.unige.aou.model.settings.License;
import ch.unige.aou.repository.DocumentFileRepository;
import ch.unige.aou.service.files.FileMetadataExtractor;
import ch.unige.aou.specification.DocumentFileSpecification;

@Service
public class DocumentFileService extends AouResourceService<DocumentFile> {
  private static final Logger log = LoggerFactory.getLogger(DocumentFileService.class);
  private static final String EMBARGO_MISSING_INFO = "documentFile.embargo.error.missing_info";
  private static final String EMBARGO_BAD_ACCESS_LEVEL = "documentFile.embargo.error.bad_access_level";

  private final DocumentFileTypeService documentFileTypeService;
  private final PublicationService publicationService;
  private final LicenseService licenseService;
  private final FileMetadataExtractor fileMetadataExtractor;

  private String tmpPublicationsFolder;

  public DocumentFileService(AouProperties aouProperties, DocumentFileTypeService documentFileTypeService, PublicationService publicationService,
          LicenseService licenseService, FileMetadataExtractor fileMetadataExtractor) {
    this.documentFileTypeService = documentFileTypeService;
    this.publicationService = publicationService;
    this.licenseService = licenseService;
    this.fileMetadataExtractor = fileMetadataExtractor;

    this.tmpPublicationsFolder = aouProperties.getTempLocation(aouProperties.getPublicationsFolder());
  }

  @Override
  protected void beforeValidate(DocumentFile documentFile) {
    this.setDefaultSortValue(documentFile);

    // forces Access level to private for Agreement files
    if (documentFile.getDocumentFileType() != null) {
      String documentFileTypeValue = documentFile.getDocumentFileType().getValue();
      if (StringTool.isNullOrEmpty(documentFileTypeValue)) {
        documentFileTypeValue = this.documentFileTypeService.findOne(documentFile.getDocumentFileType().getResId()).getValue();
      }

      if (DocumentFileType.AGREEMENT_VALUE.equals(documentFileTypeValue) ||
              DocumentFileType.IMPRIMATUR_VALUE.equals(documentFileTypeValue) ||
              DocumentFileType.PUBLICATION_MODE_VALUE.equals(documentFileTypeValue)) {
        documentFile.setAccessLevel(DocumentFile.AccessLevel.PRIVATE);
        documentFile.setEmbargoAccessLevel(null);
        documentFile.setEmbargoEndDate(null);
      }
    }
  }

  @Override
  protected void validateItemSpecificRules(DocumentFile documentFile, BindingResult errors) {

    /**
     * Check eventual embargo values coherence
     */
    if (documentFile.getEmbargoAccessLevel() != null || documentFile.getEmbargoEndDate() != null) {

      if (documentFile.getEmbargoAccessLevel() == null) {
        errors.addError(
                new FieldError(documentFile.getClass().getSimpleName(), "embargoAccessLevel", this.messageService.get(EMBARGO_MISSING_INFO)));
      } else if (documentFile.getEmbargoEndDate() == null) {
        errors.addError(
                new FieldError(documentFile.getClass().getSimpleName(), "embargoEndDate", this.messageService.get(EMBARGO_MISSING_INFO)));
      } else {
        // check that the final access level is more open that the embargo access level
        if (documentFile.getAccessLevel() != null && documentFile.getAccessLevel().value() >= documentFile.getEmbargoAccessLevel().value()) {
          errors.addError(new FieldError(documentFile.getClass().getSimpleName(), "embargoAccessLevel", this.messageService
                  .get(EMBARGO_BAD_ACCESS_LEVEL,
                          new Object[] { documentFile.getEmbargoAccessLevel().name(), documentFile.getAccessLevel().name() })));
        }
      }
    }
  }

  @Transactional
  public DocumentFile uploadNewDocumentFile(String publicationId, String documentFileTypeId, DocumentFile.AccessLevel accessLevel,
          DocumentFile.AccessLevel embargoAccessLevel,
          LocalDate embargoEndDate,
          String licenseId,
          String label,
          MultipartFile multipartFile) {

    Publication publication = this.publicationService.findOne(publicationId);
    DocumentFileType documentFileType = this.documentFileTypeService.findOne(documentFileTypeId);

    DocumentFile documentFile = new DocumentFile();
    documentFile.setPublication(publication);
    documentFile.setDocumentFileType(documentFileType);
    documentFile.setAccessLevel(accessLevel);

    if (embargoAccessLevel != null) {
      documentFile.setEmbargoAccessLevel(embargoAccessLevel);
    }

    if (embargoEndDate != null) {
      documentFile.setEmbargoEndDate(embargoEndDate);
    }

    if (!StringTool.isNullOrEmpty(licenseId)) {
      License license = this.licenseService.findOne(licenseId);
      documentFile.setLicense(license);
    }

    if (!StringTool.isNullOrEmpty(label)) {
      documentFile.setLabel(label);
    }

    // copy uploaded file to tmp folder and set sourceData
    try {
      Path tmpFolder = Paths.get(this.tmpPublicationsFolder).resolve(publicationId);
      if (FileTool.ensureFolderExists(tmpFolder)) {
        Path tmpFile = tmpFolder.resolve(multipartFile.getOriginalFilename());
        multipartFile.transferTo(tmpFile);
        documentFile.setSourceData(tmpFile.toUri());
      } else {
        throw new SolidifyRuntimeException("Could not create tmp folder: " + tmpFolder);
      }
    } catch (IOException ioe) {
      throw new SolidifyRuntimeException("Could not save uploaded file: " + multipartFile.getOriginalFilename(), ioe);
    }

    return this.save(documentFile);
  }

  @Override
  public SolidifySpecification<DocumentFile> getSpecification(DocumentFile resource) {
    return new DocumentFileSpecification(resource);
  }

  public void checkFileValidity(DocumentFile documentFile) {
    try {
      if (MediaType.APPLICATION_PDF_VALUE.equals(documentFile.getMimetype())) {
        this.checkPdfValidity(documentFile);
      }
    } catch (IOException io) {
      documentFile.setStatus(DocumentFile.DocumentFileStatus.TO_BE_CONFIRMED);
      documentFile.setStatusMessage(this.messageService.get("documentFile.status.to_be_confirmed.parsing_error"));
      log.error("An error occurred while checking validity of file '{}' for deposit {}", documentFile.getFinalData(),
              documentFile.getPublication().getResId(), io);
    }
  }

  private void checkPdfValidity(DocumentFile documentFile) throws IOException {

    /**
     * Check if the document file can be read
     */
    if (!this.fileMetadataExtractor.isValidFile(documentFile)) {
      documentFile.setStatus(DocumentFile.DocumentFileStatus.IMPOSSIBLE_TO_DOWNLOAD);
      documentFile.setStatusMessage(this.messageService.get("documentFile.status.impossible_to_download.not_valid_pdf"));
      log.warn("The document file '{}' for the deposit '{}' seems to be invalid", documentFile.getFinalData(),
              documentFile.getPublication().getResId());
    } else if (this.fileMetadataExtractor.isPasswordProtected(documentFile)) {

      /**
       * Check if the document is password protected
       */
      documentFile.setStatus(DocumentFile.DocumentFileStatus.TO_BE_CONFIRMED);
      documentFile.setStatusMessage(this.messageService.get("documentFile.status.to_be_confirmed.password_protected"));
      log.error("Unable to read password protected PDF file '{}' for deposit {}", documentFile.getFinalData(),
              documentFile.getPublication().getResId());
    } else {

      /**
       * Check that some text can be extracted from file
       */
      String text = this.fileMetadataExtractor.getText(documentFile);
      if (text != null) {
        text = text.trim();
      }
      if (StringUtils.isEmpty(text)) {
        documentFile.setStatus(DocumentFile.DocumentFileStatus.TO_BE_CONFIRMED);
        documentFile.setStatusMessage(this.messageService.get("documentFile.status.to_be_confirmed.empty_text"));
        log.warn("The document file '{}' for the deposit '{}' does not seem to contain any text", documentFile.getFinalData(),
                documentFile.getPublication().getResId());
      }
    }
  }

  public Map<DocumentFileStatus, Long> getMapStatusOccurences(String publicationId) {
    List<DocumentFileStatisticsDTO> listStatistics = ((DocumentFileRepository) this.itemRepository).getStatusStatistics(publicationId);
    return DocumentFileStatisticsDTO.getMapFromList(listStatistics);
  }

  public boolean statusUpdateIsValid(DocumentFile documentFile, DocumentFileStatus newStatus) {
    if (documentFile.getStatus() == DocumentFileStatus.TO_BE_CONFIRMED) {
      return newStatus == DocumentFileStatus.CONFIRMED;
    }
    return false;
  }

  public boolean updateSortValues(String parentId, List<String> ids) {
    int index = 0;
    for (String documentFileId : ids) {
      DocumentFile documentFile = this.findOne(documentFileId);
      if (documentFile.getPublication().getResId().equals(parentId)) {
        int newSortValue = (index + 1) * AouConstants.ORDER_INCREMENT; // convert sortValues to 10, 20, 30, 40
        documentFile.setSortValue(newSortValue);
        this.save(documentFile);
      } else {
        return false; // the publication id associated with documentFile didn't match with the one provided
      }
      index++;
    }
    return true;
  }

  @Override
  public Optional<Integer> getMaxSortValue(Sortable item) {
    return ((DocumentFileRepository) this.itemRepository).findMaxSortValueByPublication(((DocumentFile) item).getPublication().getResId());
  }

  @Override
  public void delete(String resId) {
    DocumentFile documentFile = this.findOne(resId);
    super.delete(resId);

    // final data may be null if the file could not be downloaded
    if (documentFile.getFinalData() != null) {
      try {
        Files.deleteIfExists(Path.of(documentFile.getFinalData()));
      } catch (IOException e) {
        throw new SolidifyFileDeleteException(e.getMessage());
      }
    }
  }

  public void verifyIfThereAreStillDuplicates(String publicationId) {
    //verify if there are still duplicate files
    Publication publication = this.publicationService.findOne(publicationId);
    //change duplicated status to ready if there is no duplicate for that document file
    publication.getDocumentFiles().stream().filter(df -> df.getStatus().equals(DocumentFile.DocumentFileStatus.DUPLICATE))
            .map(df -> {
              if (!this.hasDuplicateFiles(df)) {
                df.setStatus(DocumentFile.DocumentFileStatus.READY);
                this.save(df);
              }
              return true;
            }).collect(Collectors.toList());

  }

  public boolean hasDuplicateFiles(DocumentFile documentFile) {
    Publication publication = documentFile.getPublication();
    if (publication.getDocumentFiles().size() > 1) {
      String checksumToVerify = documentFile.getChecksum();
      for (int i = 0; i < publication.getDocumentFiles().size(); i++) {
        if (checksumToVerify.equals(publication.getDocumentFiles().get(i).getChecksum())
                && !publication.getDocumentFiles().get(i).equals(documentFile)) {
          return true;
        }
      }
    }
    return false;
  }
}
