package ch.unige.aou.business;

import static ch.unige.aou.AouConstants.NO_REPLY_PREFIX;

import java.security.Principal;
import java.util.List;
import java.util.NoSuchElementException;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.web.client.RestTemplate;

import ch.unige.solidify.auth.client.service.AuthorizationClientProperties;
import ch.unige.solidify.auth.client.service.UserCreatorService;
import ch.unige.solidify.auth.model.AuthApplicationRole;
import ch.unige.solidify.auth.model.AuthUser;
import ch.unige.solidify.model.SolidifyApplicationRole;
import ch.unige.solidify.service.HttpRequestInfoProvider;
import ch.unige.solidify.util.PropagateRestClientTool;
import ch.unige.solidify.util.TrustedRestClientTool;

import ch.unige.aou.AouConstants;
import ch.unige.aou.config.AouProperties;
import ch.unige.aou.model.notification.NotificationType;
import ch.unige.aou.model.security.User;
import ch.unige.aou.model.settings.Person;
import ch.unige.aou.repository.PersonRepository;
import ch.unige.aou.repository.UserRepository;
import ch.unige.aou.specification.UserSpecification;

@Service
public class UserService extends AouResourceService<User> implements UserCreatorService {
  static final Logger log = LoggerFactory.getLogger(UserService.class);

  private static final String USERS_PATH = "/users/";
  private static final String DELETE_TOKEN_PATH = "/delete-token";
  private static final String CHANGE_ROLE_PATH = "/change-role";

  private final UserRepository userRepository;
  private final PersonRepository personRepository;
  private final NotificationTypeService notificationTypeService;
  private final HttpRequestInfoProvider httpRequestInfoProvider;
  private final TrustedRestClientTool trustedRestClientTool;
  private final PropagateRestClientTool propagateRestClientTool;

  private final String authModuleUrl;

  public UserService(UserRepository userRepository, PersonRepository personRepository,
          NotificationTypeService notificationTypeService, TrustedRestClientTool trustedRestClientTool,
          HttpRequestInfoProvider httpRequestInfoProvider, AuthorizationClientProperties authProperties,
          PropagateRestClientTool propagateRestClientTool) {
    this.userRepository = userRepository;
    this.personRepository = personRepository;
    this.notificationTypeService = notificationTypeService;
    this.httpRequestInfoProvider = httpRequestInfoProvider;
    this.trustedRestClientTool = trustedRestClientTool;
    this.propagateRestClientTool = propagateRestClientTool;
    this.authModuleUrl = authProperties.getAuthorizationServerUrl();
  }

  @Override
  public UserSpecification getSpecification(User resource) {
    return new UserSpecification(resource);
  }

  public User findByExternalUid(String uniqueId) {
    return this.userRepository.findByExternalUid(uniqueId).orElseThrow(() -> new NoSuchElementException("No user with external id " + uniqueId));
  }

  public User findByPersonResId(String personId) {
    return this.userRepository.findByPersonResId(personId);
  }

  public List<User> findAdminOrRootUsers() {
    return this.userRepository.findAdminOrRootUsers();
  }

  public void revokeAccessAndRefreshTokens(User user) {
    RestTemplate restTemplate = this.trustedRestClientTool.getClient();
    restTemplate.delete(this.authModuleUrl + USERS_PATH + user.getExternalUid() + DELETE_TOKEN_PATH);
  }

  @Override
  public User save(User user) {
    if (user.getApplicationRole() == null) {
      user.setApplicationRole(new SolidifyApplicationRole(AuthApplicationRole.USER));
    }
    return super.save(user);
  }

  public void createTestUser(AouProperties.Test.User user, Person person) {
    String name = user.getName().toUpperCase();
    String mail = NO_REPLY_PREFIX + name + "@unige.ch";
    User userDb = this.userRepository.findByEmail(mail);
    if (userDb == null) {
      User newUser = new User();
      newUser.setResId(user.getName().toLowerCase());
      newUser.setFirstName(AouConstants.PERMANENT_TEST_DATA_LABEL + name + "_firstName");
      newUser.setLastName(AouConstants.PERMANENT_TEST_DATA_LABEL + name + "_lastName");
      newUser.setEmail(mail);
      newUser.setHomeOrganization(name + "_homeOrganization");
      newUser.setExternalUid(name + "_externalUid");
      newUser.setPerson(person);
      if (user.getRole().equalsIgnoreCase(AuthApplicationRole.ADMIN.getName())) {
        newUser.setApplicationRole(new SolidifyApplicationRole(AuthApplicationRole.ADMIN));
      } else if (user.getRole().equalsIgnoreCase(AuthApplicationRole.ROOT.getName())) {
        newUser.setApplicationRole(new SolidifyApplicationRole(AuthApplicationRole.ROOT));
      } else if (user.getRole().equalsIgnoreCase(AuthApplicationRole.TRUSTED_CLIENT.getName())) {
        newUser.setApplicationRole(new SolidifyApplicationRole(AuthApplicationRole.TRUSTED_CLIENT));
      } else {
        newUser.setApplicationRole(new SolidifyApplicationRole(AuthApplicationRole.USER));
      }
      User createdUser = this.save(newUser);
      log.info("Test user '{}' created", createdUser.getResId());
    }
  }

  @Override
  @Transactional
  public synchronized void createNewUser(String externalUid) {
    if (!this.userRepository.existsByExternalUid(externalUid)) {
      final AuthUser authUser = this.getAuthenticatedUser(externalUid);
      // Create a new user
      User newUser = new User(authUser);
      // Update Last Login
      this.updateLastLogin(newUser, authUser);
      // Create Person
      Person newPerson = new Person();
      newPerson.setFirstName(newUser.getFirstName());
      newPerson.setLastName(newUser.getLastName());
      this.addNotificationType(newPerson, NotificationType.MY_PUBLICATION_VALIDATED.getResId());
      this.addNotificationType(newPerson, NotificationType.MY_INDIRECT_PUBLICATION_VALIDATED.getResId());
      this.personRepository.save(newPerson);
      newUser.setPerson(newPerson);
      // Save user
      this.userRepository.save(newUser);
    }
  }

  private void addNotificationType(Person person, String notificationTypeId) {
    NotificationType notificationType = this.notificationTypeService.findOne(notificationTypeId);
    person.addNotificationType(notificationType);
  }

  public User getCurrentUser() {
    final Principal principal = this.httpRequestInfoProvider.getPrincipal();
    return this.findByExternalUid(principal.getName());
  }

  public void updateUser(User user) {
    final AuthUser authUser = this.getAuthenticatedUser(user.getExternalUid());
    this.updateLastLogin(user, authUser);
    this.updateApplicationRole(user, authUser);
  }

  private AuthUser getAuthenticatedUser(String externalUid) {
    RestTemplate restTemplate = this.trustedRestClientTool.getClient();
    AuthUser authUser = restTemplate.getForEntity(this.authModuleUrl + USERS_PATH + externalUid, AuthUser.class).getBody();
    if (authUser == null) {
      throw new IllegalStateException();
    }
    return authUser;
  }

  public AuthUser updateUserRole(String externalUid, String role) {
    RestTemplate restTemplate = this.propagateRestClientTool.getClient();
    AuthUser authUser = restTemplate.postForObject(this.authModuleUrl + USERS_PATH + externalUid + CHANGE_ROLE_PATH, role, AuthUser.class);
    if (authUser == null) {
      throw new IllegalStateException();
    }
    return authUser;
  }

  private void updateLastLogin(User user, AuthUser authUser) {
    if (authUser.getLastLogin() != null) {
      user.setLastLoginIpAddress(authUser.getLastLogin().getLoginIpAddress());
      user.setLastLoginTime(authUser.getLastLogin().getLoginTime());
    } else if (authUser.getFirstLogin() != null) {
      user.setLastLoginIpAddress(authUser.getFirstLogin().getLoginIpAddress());
      user.setLastLoginTime(authUser.getFirstLogin().getLoginTime());
    }
  }

  private void updateApplicationRole(User user, AuthUser authUser) {
    user.setApplicationRole(new SolidifyApplicationRole(authUser.getApplicationRole()));
  }

}
