package ch.unige.aou.service.metadata.imports;

import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.util.Base64;
import java.util.HashMap;
import java.util.Map;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpressionException;

import org.springframework.boot.web.client.RestTemplateBuilder;
import org.springframework.http.HttpStatus;
import org.springframework.stereotype.Service;
import org.springframework.web.client.HttpClientErrorException;
import org.springframework.web.client.RestTemplate;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;

import ch.unige.solidify.exception.SolidifyHttpErrorException;
import ch.unige.solidify.exception.SolidifyRuntimeException;
import ch.unige.solidify.service.MessageService;
import ch.unige.solidify.util.StringTool;
import ch.unige.solidify.util.XMLTool;

import ch.unige.aou.business.ContributorService;
import ch.unige.aou.business.DocumentFileTypeService;
import ch.unige.aou.config.AouProperties;
import ch.unige.aou.model.xml.deposit.v2_3.Contributor;
import ch.unige.aou.model.xml.deposit.v2_3.DateTypes;
import ch.unige.aou.model.xml.deposit.v2_3.DepositDoc;
import ch.unige.aou.model.xml.deposit.v2_3.Text;
import ch.unige.aou.service.MetadataService;
import ch.unige.aou.service.rest.JournalTitleRemoteService;

@Service
public class DataCiteImportService extends MetadataImportService {

  private static final String DATACITE_SCHEMA_2_2 = "http://datacite.org/schema/kernel-2.2";
  private static final String DATACITE_SCHEMA_3 = "http://datacite.org/schema/kernel-3";
  private static final String DATACITE_SCHEMA_4 = "http://datacite.org/schema/kernel-4";
  private static final String XPATH_DOI = "/d:resource/d:identifier[@identifierType=\"DOI\"]";
  private static final String XPATH_TITLE = "/d:resource/d:titles/d:title";
  private static final String XPATH_CREATORS = "/d:resource/d:creators/d:creator";
  private static final String XPATH_ABSTRACT = "/d:resource/d:descriptions/d:description[@descriptionType=\"Abstract\"]";
  private static final String XPATH_PUBLICATION_YEAR = "/d:resource/d:publicationYear";
  private static final String XPATH_LANGUAGE = "/d:resource/d:language";
  private static final String XPATH_SUBJECTS = "/d:resource/d:subjects/d:subject";

  private final String apiUrl;

  public DataCiteImportService(AouProperties aouProperties,
          MessageService messageService,
          MetadataService metadataService,
          ContributorService contributorService,
          DocumentFileTypeService documentFileTypeService,
          JournalTitleRemoteService journalTitleRemoteService) {
    super(aouProperties, messageService, metadataService, contributorService, documentFileTypeService, journalTitleRemoteService);
    this.apiUrl = aouProperties.getMetadata().getImports().getDataCite().getUrl();
  }

  public DepositDoc searchDoi(String doi, DepositDoc depositDoc) {
    try {
      final JsonNode parentNode = this.getJsonFromDataCite(doi);
      final Document dataCiteDoc = this.getXmlFromJson(parentNode);
      final String schemaVersion = this.getSchemaVersionFromJson(parentNode);
      this.fillDepositDoc(doi, depositDoc, dataCiteDoc, schemaVersion);
      return depositDoc;
    } catch (final XPathExpressionException e) {
      throw new SolidifyRuntimeException("Unable to search DOI by Datacite", e);
    }
  }

  public void fillDepositDoc(String doi, DepositDoc depositDoc, Document dataciteDoc, String schemaVersion) throws XPathExpressionException {
    final Map<String, String> namespace;
    switch (schemaVersion) {
      case "4":
        namespace = Map.of("d", DATACITE_SCHEMA_4);
        break;
      case "3":
        namespace = Map.of("d", DATACITE_SCHEMA_3);
        break;
      case "2.2":
        namespace = Map.of("d", DATACITE_SCHEMA_2_2);
        break;
      default:
        throw new SolidifyRuntimeException("Datacite version: " + schemaVersion + " not supported");
    }
    final XPath xpath = XMLTool.buildXPath(namespace);

    Node doiNode = (Node) xpath.compile(XPATH_DOI).evaluate(dataciteDoc, XPathConstants.NODE);
    Node titleNode = (Node) xpath.compile(XPATH_TITLE).evaluate(dataciteDoc, XPathConstants.NODE);
    Node creatorNode = (Node) xpath.compile(XPATH_CREATORS).evaluate(dataciteDoc, XPathConstants.NODE);
    Node abstractNode = (Node) xpath.compile(XPATH_ABSTRACT).evaluate(dataciteDoc, XPathConstants.NODE);
    Node publicationYearNode = (Node) xpath.compile(XPATH_PUBLICATION_YEAR).evaluate(dataciteDoc, XPathConstants.NODE);
    Node languageNode = (Node) xpath.compile(XPATH_LANGUAGE).evaluate(dataciteDoc, XPathConstants.NODE);
    Node subjectNode = (Node) xpath.compile(XPATH_SUBJECTS).evaluate(dataciteDoc, XPathConstants.NODE);

    if (doiNode != null) {
      this.fillDoi(depositDoc, dataciteDoc, XPATH_DOI, namespace);
    }
    if (titleNode != null) {
      this.fillTitle(depositDoc, dataciteDoc, XPATH_TITLE, namespace);
    }
    if (creatorNode != null) {
      this.fillContributors(depositDoc, dataciteDoc, XPATH_CREATORS, namespace);
    }
    if (abstractNode != null) {
      this.fillAbstracts(depositDoc, dataciteDoc, XPATH_ABSTRACT, namespace);
    }
    if (publicationYearNode != null) {
      this.fillPublicationYear(depositDoc, dataciteDoc, XPATH_PUBLICATION_YEAR, namespace);
    }
    if (languageNode != null) {
      this.fillLanguage(depositDoc, dataciteDoc, XPATH_LANGUAGE, namespace);
    }
    if (subjectNode != null) {
      this.fillKeywords(depositDoc, dataciteDoc, XPATH_SUBJECTS, namespace);
    }

    // if the DOI fetched from metadata is different from the one used to retrieve metadata, replace it
    this.ensureDepositIdentifiersExist(depositDoc);
    if (StringTool.isNullOrEmpty(depositDoc.getIdentifiers().getDoi()) || !depositDoc.getIdentifiers().getDoi().equals(doi)) {
      depositDoc.getIdentifiers().setDoi(doi);
    }
  }

  private JsonNode getJsonFromDataCite(String doi) {
    try {
      RestTemplateBuilder restTemplateBuilder = new RestTemplateBuilder();
      RestTemplate client = restTemplateBuilder.rootUri(this.apiUrl).build();
      Map<String, String> parameters = new HashMap<>();
      parameters.put("doi", doi);
      final String jsonString = client.getForObject(this.apiUrl, String.class, parameters);
      return new ObjectMapper().readTree(jsonString);
    } catch (HttpClientErrorException e) {
      throw new SolidifyHttpErrorException(HttpStatus.NOT_FOUND,
              this.messageService.get("deposit.error.identifiers.doi.not_found_by_datacite", new Object[] { doi }));
    } catch (IOException e) {
      throw new SolidifyRuntimeException("Unable to search DOI by DataCite", e);
    }
  }

  private Document getXmlFromJson(JsonNode parentNode) {
    final String encodedXmlString = parentNode.get("data").get("attributes").get("xml").textValue();
    final String xmlString = new String(Base64.getDecoder().decode(encodedXmlString), StandardCharsets.UTF_8);
    try {
      return XMLTool.parseXML(xmlString);
    } catch (ParserConfigurationException | SAXException | IOException e) {
      throw new SolidifyRuntimeException("Unable to search DOI by DataCite", e);
    }
  }

  private String getSchemaVersionFromJson(JsonNode parentNode) {
    return parentNode.get("data").get("attributes").get("schema-version").textValue();
  }

  private void fillAbstracts(DepositDoc depositDoc, Document dataciteDoc, String xpathQuery, Map<String, String> namespaces) {
    String value = XMLTool.getFirstNodeInnerRawContent(dataciteDoc, xpathQuery, namespaces);

    if (!StringTool.isNullOrEmpty(value)) {
      this.ensureAbstractsExist(depositDoc);
      Text text = this.getText(value);
      depositDoc.getAbstracts().getAbstract().add(text);
    }
  }

  private void fillContributors(DepositDoc depositDoc, Document dataciteDoc, String xpathQuery, Map<String, String> namespaces) {
    final XPath xpath = XMLTool.buildXPath(namespaces);
    try {
      NodeList authorsList = (NodeList) xpath.compile(xpathQuery).evaluate(dataciteDoc, XPathConstants.NODESET);

      if (authorsList.getLength() > 0) {
        this.ensureContributorsExist(depositDoc);
      }

      for (int i = 0; i < this.getNumberMaxContributor(authorsList.getLength()); i++) {
        String familyName = XMLTool.getFirstTextContent(authorsList.item(i), "./d:familyName", namespaces);
        String givenName = XMLTool.getFirstTextContent(authorsList.item(i), "./d:givenName", namespaces);

        if (familyName == null && givenName == null) {
          String creatorName = XMLTool.getFirstTextContent(authorsList.item(i), "./d:creatorName", namespaces);
          if (creatorName != null && creatorName.contains(",")) {
            familyName = creatorName.split(",")[0].trim();
            givenName = creatorName.split(",")[1].trim();
          }
        }

        familyName = this.getNameOrUndefinedValue(familyName);
        givenName = this.getNameOrUndefinedValue(givenName);

        Contributor contributor = new Contributor();
        contributor.setLastname(familyName);
        contributor.setFirstname(givenName);
        depositDoc.getContributors().getContributorOrCollaboration().add(contributor);
      }
    } catch (XPathExpressionException e) {
      throw new SolidifyRuntimeException("Unable to fill contributors", e);
    }
  }

  private void fillPublicationYear(DepositDoc depositDoc, Document dataciteDoc, String xpathQuery, Map<String, String> namespaces) {
    String value = XMLTool.getFirstNodeInnerRawContent(dataciteDoc, xpathQuery, namespaces);

    if (!StringTool.isNullOrEmpty(value)) {
      this.ensureDatesExist(depositDoc);
      depositDoc.getDates().getDate().add(this.getDateWithType(DateTypes.PUBLICATION, value));
    }
  }

  private void fillLanguage(DepositDoc depositDoc, Document dataciteDoc, String xpathQuery, Map<String, String> namespaces) {
    String value = XMLTool.getFirstTextContent(dataciteDoc, xpathQuery, namespaces);
    if (!StringTool.isNullOrEmpty(value)) {
      this.ensureLanguagesExist(depositDoc);
      depositDoc.getLanguages().getLanguage().add(this.getLanguageCode(value));
      if (depositDoc.getTitle() != null) {
        depositDoc.getTitle().setLang(this.getLanguageCode(value));
      }
    }
  }

  private void fillKeywords(DepositDoc depositDoc, Document dataciteDoc, String xpathQuery, Map<String, String> namespaces) {
    final XPath xpath = XMLTool.buildXPath(namespaces);
    try {
      NodeList keywordsList = (NodeList) xpath.compile(xpathQuery).evaluate(dataciteDoc, XPathConstants.NODESET);

      if (keywordsList.getLength() > 0) {
        this.ensureKeywordsExist(depositDoc);
      }

      for (int i = 0; i < keywordsList.getLength(); i++) {
        String keyword = keywordsList.item(i).getTextContent();
        depositDoc.getKeywords().getKeyword().add(keyword);
      }

    } catch (XPathExpressionException e) {
      throw new SolidifyRuntimeException("Unable to fill keywords", e);
    }
  }
}
