package ch.unige.aou.service.metadata.imports;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.stream.Collectors;
import java.util.stream.StreamSupport;

import org.json.JSONArray;
import org.json.JSONObject;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.boot.web.client.RestTemplateBuilder;
import org.springframework.http.HttpStatus;
import org.springframework.stereotype.Service;
import org.springframework.web.client.HttpClientErrorException;
import org.springframework.web.client.RestTemplate;

import ch.unige.solidify.exception.SolidifyHttpErrorException;
import ch.unige.solidify.exception.SolidifyRuntimeException;
import ch.unige.solidify.service.MessageService;
import ch.unige.solidify.util.StringTool;

import ch.unige.aou.AouConstants;
import ch.unige.aou.business.ContributorService;
import ch.unige.aou.business.DocumentFileTypeService;
import ch.unige.aou.business.PublicationSubtypeService;
import ch.unige.aou.config.AouProperties;
import ch.unige.aou.model.publication.PublicationSubtype;
import ch.unige.aou.model.xml.deposit.v2_3.Contributor;
import ch.unige.aou.model.xml.deposit.v2_3.DateTypes;
import ch.unige.aou.model.xml.deposit.v2_3.DateWithType;
import ch.unige.aou.model.xml.deposit.v2_3.DepositDoc;
import ch.unige.aou.service.MetadataService;
import ch.unige.aou.service.rest.JournalTitleRemoteService;

@Service
public class InspireHepImportService extends MetadataImportService {
  private static final Logger log = LoggerFactory.getLogger(InspireHepImportService.class);
  private String apiArxivUrl;
  private String apiDoiUrl;

  protected PublicationSubtypeService publicationSubTypeService;

  public InspireHepImportService(AouProperties aouProperties, MessageService messageService,
          MetadataService metadataService, ContributorService contributorService,
          DocumentFileTypeService documentFileTypeService, JournalTitleRemoteService journalTitleRemoteService,
          PublicationSubtypeService publicationSubTypeService) {
    super(aouProperties, messageService, metadataService, contributorService, documentFileTypeService, journalTitleRemoteService);
    this.apiArxivUrl = aouProperties.getMetadata().getImports().getInspireHep().getArxivUrl();
    this.apiDoiUrl = aouProperties.getMetadata().getImports().getInspireHep().getDoiUrl();
    this.publicationSubTypeService = publicationSubTypeService;

  }

  public DepositDoc searchInspireHep(String identifier, boolean isArxivIdentifier, DepositDoc depositDoc) {
    if (!StringTool.isNullOrEmpty(this.apiArxivUrl) && !StringTool.isNullOrEmpty(this.apiDoiUrl)) {
      try {
        JSONObject inspireHepDoc = this.getJsonFromInspireHep(identifier, isArxivIdentifier);
        this.fillDepositDoc(depositDoc, inspireHepDoc);

        return depositDoc;
      } catch (Exception e) {
        throw new SolidifyRuntimeException("Unable to search Archive by InspireHep", e);
      }
    } else {
      throw new SolidifyRuntimeException("Incomplete configuration");
    }
  }

  private JSONObject getJsonFromInspireHep(String identifier, boolean isArxivIdentifier) {
    try {
      RestTemplateBuilder restTemplateBuilder = new RestTemplateBuilder();
      RestTemplate client = restTemplateBuilder.rootUri(this.apiArxivUrl).build();

      Map<String, String> parameters = new HashMap<>();
      final String jsonString;
      if (isArxivIdentifier) {
        parameters.put("arxivId", identifier);
        jsonString = client.getForObject(this.apiArxivUrl, String.class, parameters);
      } else {
        parameters.put("doi", identifier);
        jsonString = client.getForObject(this.apiDoiUrl, String.class, parameters);
      }

      return new JSONObject(jsonString);
    } catch (HttpClientErrorException e) {
      if (isArxivIdentifier) {
        throw new SolidifyHttpErrorException(HttpStatus.NOT_FOUND,
                this.messageService.get("deposit.error.identifiers.arxiv.not_found_by_inspire_hep", new Object[] { identifier }));
      } else {
        throw new SolidifyHttpErrorException(HttpStatus.NOT_FOUND,
                this.messageService.get("deposit.error.identifiers.doi.not_found_by_inspire_hep", new Object[] { identifier }));
      }
    }
  }

  /****************************************************title************************************/

  public void fillDepositDoc(DepositDoc depositDoc, JSONObject inspireHepJson) {
    //check first if the metadata curated property is true to continue getting information from InspireHep
    if (inspireHepJson.getJSONObject("metadata").getBoolean("curated")) {
      JSONObject metadata = inspireHepJson.getJSONObject("metadata");
      this.fillTitle(depositDoc, this.getArrayOrNull(metadata, "titles"));
      this.fillPublicationType(depositDoc, this.getArrayOrNull(metadata, "document_type"));
      this.fillAbstracts(depositDoc, this.getArrayOrNull(metadata, "abstracts"));
      this.fillPublicationDate(depositDoc, this.getArrayOrNull(metadata, "publication_info"));
      this.fillDoi(depositDoc, this.getArrayOrNull(metadata, "dois"));
      this.fillContainerTitle(depositDoc, this.getArrayOrNull(metadata, "publication_info"));
      this.fillContainerVolume(depositDoc, this.getArrayOrNull(metadata, "publication_info"));
      this.fillContributors(depositDoc, this.getArrayOrNull(metadata, "authors"));
      this.fillKeywords(depositDoc, this.getArrayOrNull(metadata, "keywords"));
      this.fillNote(depositDoc, this.getArrayOrNull(metadata, "public_notes"));
      this.fillArxivId(depositDoc, this.getArrayOrNull(metadata, "arxiv_eprints"));
    } else {
      throw new SolidifyRuntimeException("The metadata found are not curated");
    }

  }

  private void fillPublicationType(DepositDoc depositDoc, JSONArray jsonArray) {
    if (jsonArray != null && !jsonArray.isEmpty()) {
      String subType = this.transformInspireHepSubType(jsonArray.getString(0));
      depositDoc.setSubtype(subType);
      Optional<PublicationSubtype> publicationSubtypeOpt = this.publicationSubTypeService.findByName(subType);
      if (publicationSubtypeOpt.isPresent()) {
        depositDoc.setType(publicationSubtypeOpt.get().getPublicationType().getName());
      }
    }
  }

  private void fillArxivId(DepositDoc depositDoc, JSONArray jsonArray) {
    if (jsonArray != null && !jsonArray.isEmpty()) {
      String value = this.getStringOrNull(jsonArray.getJSONObject(0), "value");
      if (!StringTool.isNullOrEmpty(value)) {
        this.ensureDepositIdentifiersExist(depositDoc);
        depositDoc.getIdentifiers().setArxiv(value);
      }
    }
  }

  private void fillPublicationDate(DepositDoc depositDoc, JSONArray jsonArray) {
    if (jsonArray != null && !jsonArray.isEmpty()) {
      String value = jsonArray.getJSONObject(0).get("year").toString();
      if (!StringTool.isNullOrEmpty(value)) {
        this.ensureDatesExist(depositDoc);
        DateWithType date = new DateWithType();
        date.setType(DateTypes.PUBLICATION);
        date.setContent(value);
        depositDoc.getDates().getDate().add(date);
      }
    }
  }

  private void fillTitle(DepositDoc depositDoc, JSONArray jsonArray) {
    if (jsonArray != null && !jsonArray.isEmpty()) {
      String value = this.getStringOrNull(jsonArray.getJSONObject(0), "title");
      if (!StringTool.isNullOrEmpty(value)) {
        value = this.cleanText(value);
        depositDoc.setTitle(this.getText(value));
      }
    }
  }

  private void fillAbstracts(DepositDoc depositDoc, JSONArray jsonArray) {
    if (jsonArray != null && !jsonArray.isEmpty()) {
      this.ensureAbstractsExist(depositDoc);
      for (int i = 0; i < jsonArray.length(); i++) {
        String value = this.getStringOrNull(jsonArray.getJSONObject(i), "value");
        if (!StringTool.isNullOrEmpty(value)) {
          value = this.cleanText(value);
          depositDoc.getAbstracts().getAbstract().add(this.getText(value));
        }
      }
    }
  }

  private void fillDoi(DepositDoc depositDoc, JSONArray jsonArray) {
    if (jsonArray != null && !jsonArray.isEmpty()) {
      String value = jsonArray.getJSONObject(0).get("value").toString();
      if (!StringTool.isNullOrEmpty(value)) {
        this.ensureDepositIdentifiersExist(depositDoc);
        depositDoc.getIdentifiers().setDoi(value);
      }
    }
  }

  private void fillContainerTitle(DepositDoc depositDoc, JSONArray jsonArray) {
    if (jsonArray != null && !jsonArray.isEmpty()) {
      String journalTitle = this.getStringOrNull(jsonArray.getJSONObject(0), "journal_title");
      if (!StringTool.isNullOrEmpty(journalTitle)) {
        this.ensureContainerExist(depositDoc);
        depositDoc.getContainer().setTitle(journalTitle);
      }
    }
  }

  private void fillContainerVolume(DepositDoc depositDoc, JSONArray jsonArray) {
    if (jsonArray != null && !jsonArray.isEmpty()) {
      String volume = this.getStringOrNull(jsonArray.getJSONObject(0), "journal_volume");
      if (!StringTool.isNullOrEmpty(volume)) {
        this.ensureContainerExist(depositDoc);
        depositDoc.getContainer().setVolume(volume);
      }
    }
  }

  @SuppressWarnings("java:S2259")
  // authorName.trim() cannot throw an NPE because authorName is checked by the method isNullOrEmpty
  // this false positive should be remove when https://jira.sonarsource.com/browse/SONARJAVA-3502 will be solved
  private void fillContributors(DepositDoc depositDoc, JSONArray jsonArray) {
    if (jsonArray != null && !jsonArray.isEmpty()) {
      this.ensureContributorsExist(depositDoc);
      for (int i = 0; i < this.getNumberMaxContributor(jsonArray.length()); i++) {
        Contributor contributor = this.createContributorFromJSONObject(jsonArray.getJSONObject(i));
        depositDoc.getContributors().getContributorOrCollaboration().add(contributor);
      }
      // add authors that have an affiliation with unige, geneve, or hug
      List<JSONObject> authors = StreamSupport.stream(jsonArray.spliterator(), false)
              .map(a -> (JSONObject) a)
              .collect(Collectors.toList());
      List<JSONObject> authorsWithAffiliation = authors.stream()
              .filter(obj -> {
                JSONArray affiliations = this.getArrayOrNull(obj, "affiliations");
                if (affiliations != null && affiliations.length() > 0) {
                  String institution = this.getStringOrNull(affiliations.getJSONObject(0), "value");
                  if (institution.startsWith("Geneva") || institution.startsWith("Genève") || institution.startsWith("Genéve")
                          || institution.startsWith("HUG") || institution.startsWith("Unige")) {
                    return true;
                  }
                }
                return false;
              }).collect(Collectors.toList());

      for (int i = 0; i < authorsWithAffiliation.size(); i++) {
        Contributor contributor = this.createContributorFromJSONObject(authorsWithAffiliation.get(i));
        //check if it was already added
        if (depositDoc.getContributors().getContributorOrCollaboration().stream()
                .noneMatch(c -> ((Contributor) c).getLastname().equals(contributor.getLastname())
                        && ((Contributor) c).getFirstname().equals(contributor.getFirstname()) && ((Contributor) c).getInstitution()
                        .equals(contributor.getInstitution()))) {
          depositDoc.getContributors().getContributorOrCollaboration().add(contributor);
        }
      }
    }
  }

  private void fillKeywords(DepositDoc depositDoc, JSONArray jsonArray) {
    if (jsonArray != null && !jsonArray.isEmpty()) {
      this.ensureKeywordsExist(depositDoc);
      for (int i = 0; i < jsonArray.length(); i++) {
        String value = this.getStringOrNull(jsonArray.getJSONObject(i), "value");
        if (!StringTool.isNullOrEmpty(value)) {
          depositDoc.getKeywords().getKeyword().add(value);
        }
      }
    }
  }

  private void fillNote(DepositDoc depositDoc, JSONArray jsonArray) {
    if (jsonArray != null && !jsonArray.isEmpty()) {
      String value = this.getStringOrNull(jsonArray.getJSONObject(0), "value");
      depositDoc.setNote(value);
    }
  }

  private String transformInspireHepSubType(String subType) {
    switch (subType) {
      case "activity report":
      case "note":
      case "report":
        return AouConstants.DEPOSIT_SUBTYPE_RAPPORT_DE_RECHERCHE_NAME;
      case "article":
        return AouConstants.DEPOSIT_SUBTYPE_ARTICLE_SCIENTIFIQUE_NAME;
      case "book":
        return AouConstants.DEPOSIT_SUBTYPE_LIVRE_NAME;
      case "book chapter":
        return AouConstants.DEPOSIT_SUBTYPE_CHAPITRE_DE_LIVRE_NAME;
      case "conference paper":
        return AouConstants.DEPOSIT_SUBTYPE_PRESENTATION_INTERVENTION_NAME;
      case "proceedings":
        return AouConstants.DEPOSIT_SUBTYPE_ACTES_DE_CONFERENCE_NAME;
      case "thesis":
        return AouConstants.DEPOSIT_SUBTYPE_THESE_NAME;
      default:
        return AouConstants.DEPOSIT_SUB_SUBTYPE_ARTICLE_NAME;
    }
  }

  private Contributor createContributorFromJSONObject(JSONObject authorObject) {
    String authorName = this.getStringOrNull(authorObject, "full_name");
    // Cut the author name on the last space, assuming the last word is the firstname
    String firstName = null;
    String lastName = null;
    if (!StringTool.isNullOrEmpty(authorName)) {
      authorName = authorName.trim();
      int commaIndex = authorName.lastIndexOf(',');
      if (commaIndex > -1) {
        lastName = authorName.substring(0, commaIndex);
        firstName = authorName.substring(commaIndex + 1);
      } else {
        lastName = authorName;
      }
    }
    firstName = this.getNameOrUndefinedValue(firstName);
    lastName = this.getNameOrUndefinedValue(lastName);

    Contributor contributor = new Contributor();
    contributor.setLastname(lastName);
    contributor.setFirstname(firstName);

    JSONArray affiliations = this.getArrayOrNull(authorObject, "affiliations");
    if (affiliations != null && affiliations.length() > 0) {
      String institution = this.getStringOrNull(affiliations.getJSONObject(0), "value");
      if (!StringTool.isNullOrEmpty(institution)) {
        contributor.setInstitution(institution);
      }
    }
    return contributor;
  }
}
