package ch.unige.aou.service.rest;

import java.io.IOException;
import java.util.HashMap;
import java.util.Map;

import org.springframework.boot.web.client.RestTemplateBuilder;
import org.springframework.context.annotation.Lazy;
import org.springframework.http.HttpStatus;
import org.springframework.stereotype.Service;
import org.springframework.web.client.HttpClientErrorException;
import org.springframework.web.client.RestTemplate;

import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;

import ch.unige.solidify.exception.SolidifyRuntimeException;
import ch.unige.solidify.util.StringTool;

import ch.unige.aou.business.PublicationService;
import ch.unige.aou.config.AouProperties;

@Service
public class DuplicateService {

  private static final String DOI = "doi";
  private static final String PUBMED = "pubmed";
  private static final String ARXIV = "arxiv";
  private static final String ISBN = "isbn";

  private static final String FOUND_MESSAGE = "this document already exists";
  private static final String NOT_FOUND_MESSAGE = "this document does not exist";

  private String identifiersCheckUrl;

  private PublicationService publicationService;

  public DuplicateService(@Lazy PublicationService publicationService, AouProperties aouProperties) {
    this.publicationService = publicationService;
    this.identifiersCheckUrl = aouProperties.getMetadata().getSearch().getDuplicates().getIdentifiersCheckUrl();
  }

  public boolean doiAlreadyExists(String doi) {
    return this.doiAlreadyExists(doi, null);
  }

  public boolean doiAlreadyExists(String doi, String publicationId) {
    return this.publicationService.doiAlreadyUsedByAnotherPublication(publicationId, doi) || this.identifierAlreadyPublished(DOI, doi);
  }

  public boolean arxivIdAlreadyExists(String arxivId) {
    return this.identifierAlreadyPublished(ARXIV, arxivId);
  }

  public boolean pmidAlreadyExists(String pmid) {
    return this.pmidAlreadyExists(pmid, null);
  }

  public boolean pmidAlreadyExists(String pmid, String publicationId) {
    return this.publicationService.pmidAlreadyUsedByAnotherPublication(publicationId, pmid) || this.identifierAlreadyPublished(PUBMED, pmid);
  }

  public boolean isbnAlreadyExists(String isbn) {
    return this.identifierAlreadyPublished(ISBN, isbn);
  }

  private boolean identifierAlreadyPublished(String identifierType, String value) {
    if (!StringTool.isNullOrEmpty(this.identifiersCheckUrl)) {

      try {
        RestTemplateBuilder restTemplateBuilder = new RestTemplateBuilder();
        RestTemplate client = restTemplateBuilder.rootUri(this.identifiersCheckUrl).build();

        Map<String, String> uriVariables = new HashMap<>();
        uriVariables.put("identifierType", identifierType);
        uriVariables.put("value", value);
        final String jsonResponse = client.getForObject(this.identifiersCheckUrl, String.class, uriVariables);

        // HTTP 200 --> identifier already exists
        this.checkFoundMessage(jsonResponse);
        return true;

      } catch (final HttpClientErrorException e) {
        if (e.getStatusCode() == HttpStatus.NOT_FOUND) {
          // HTTP 404 --> id not deposited yet
          this.checkNotFoundMessage(e.getResponseBodyAsString());
          return false;
        } else {
          throw new SolidifyRuntimeException(e.getMessage());
        }
      }
    } else {
      throw new SolidifyRuntimeException("incomplete configuration: identifiersCheckUrl is not set");
    }
  }

  private boolean checkFoundMessage(String jsonResponse) {
    try {
      ObjectMapper mapper = new ObjectMapper();
      JsonNode responseNode = mapper.readValue(jsonResponse, JsonNode.class);
      if (responseNode.has("success")) {
        JsonNode successNode = responseNode.get("success");
        String successMessage = successNode.get(0).asText();
        if (FOUND_MESSAGE.equals(successMessage)) {
          return true;
        }
      }
      throw new SolidifyRuntimeException("incoherent service response");
    } catch (IOException e) {
      throw new SolidifyRuntimeException(e.getMessage());
    }
  }

  private boolean checkNotFoundMessage(String jsonResponse) {
    try {
      ObjectMapper mapper = new ObjectMapper();
      JsonNode responseNode = mapper.readValue(jsonResponse, JsonNode.class);
      if (responseNode.has("errors")) {
        JsonNode errorsNode = responseNode.get("errors");
        String errorMessage = errorsNode.get(0).asText();
        if (NOT_FOUND_MESSAGE.equals(errorMessage)) {
          return true;
        }
      }
      throw new SolidifyRuntimeException("incoherent service response");
    } catch (IOException ex) {
      throw new SolidifyRuntimeException(ex.getMessage());
    }
  }
}
