package ch.unige.aou.service.rest;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Objects;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.boot.web.client.RestTemplateBuilder;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpHeaders;
import org.springframework.http.HttpMethod;
import org.springframework.http.HttpStatus;
import org.springframework.http.MediaType;
import org.springframework.http.ResponseEntity;
import org.springframework.scheduling.annotation.Scheduled;
import org.springframework.stereotype.Service;
import org.springframework.web.client.RestClientException;
import org.springframework.web.client.RestTemplate;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.node.ArrayNode;

import ch.unige.solidify.SolidifyConstants;
import ch.unige.solidify.exception.SolidifyRuntimeException;
import ch.unige.solidify.rest.Collection;
import ch.unige.solidify.rest.CollectionPage;
import ch.unige.solidify.util.StringTool;

import ch.unige.aou.config.AouProperties;
import ch.unige.aou.model.rest.IssnAuthInfo;
import ch.unige.aou.model.rest.JournalTitleDTO;
import ch.unige.aou.model.rest.SearchDTO;

@Service
public class JournalTitleRemoteService {
  private static final Logger log = LoggerFactory.getLogger(JournalTitleRemoteService.class);
  private static final long DOWNLOAD_TOKEN_INTERVAL = 600000; // ten minutes
  private static final String AUTHENTICATE = "authenticate";
  private static final String NOTICE = "notice";
  private static final String SEARCH = "search";
  private static final String PATH_SPATIAL = "spatial";
  private static final String PATH_ISSN = "resource/ISSN/";
  private static final String PATH_PUBLISHER = "publisher";

  private String remoteServiceUrl;
  private String user;
  private String apiKey;

  private String OAuth2Token;

  public JournalTitleRemoteService(AouProperties aouProperties) {
    this.remoteServiceUrl = aouProperties.getMetadata().getSearch().getJournal().getUrl();
    this.user = aouProperties.getMetadata().getSearch().getJournal().getUser();
    this.apiKey = aouProperties.getMetadata().getSearch().getJournal().getApiKey();
  }

  @Scheduled(fixedDelay = DOWNLOAD_TOKEN_INTERVAL)
  protected String getOAuth2TokenFromExternalService() {
    if (!StringTool.isNullOrEmpty(this.remoteServiceUrl) && !StringTool.isNullOrEmpty(this.user) && !StringTool.isNullOrEmpty(this.apiKey)) {
      String url = this.remoteServiceUrl + SolidifyConstants.URL_SEP + AUTHENTICATE + SolidifyConstants.URL_SEP + this.user
              + SolidifyConstants.URL_SEP + this.apiKey;
      RestTemplateBuilder restTemplateBuilder = new RestTemplateBuilder();
      RestTemplate client = restTemplateBuilder.rootUri(url).build();

      final IssnAuthInfo jsonString = client.getForObject(url, IssnAuthInfo.class);
      this.OAuth2Token = !Objects.isNull(jsonString) ? jsonString.getToken() : null;
      return this.OAuth2Token;
    } else {
      throw new SolidifyRuntimeException("incomplete configuration");
    }
  }

  public Collection<JournalTitleDTO> getJournalTitlesByTitle(SearchDTO searchDTO) {
    String url = this.remoteServiceUrl + SolidifyConstants.URL_SEP + SEARCH;
    if (StringTool.isNullOrEmpty(this.OAuth2Token)) {
      this.getOAuth2TokenFromExternalService();
    }

    RestTemplateBuilder restTemplateBuilder = new RestTemplateBuilder();
    RestTemplate client = restTemplateBuilder.rootUri(url).build();
    // create headers
    HttpHeaders headers = new HttpHeaders();
    headers.setContentType(MediaType.APPLICATION_JSON);
    headers.setAccept(Collections.singletonList(MediaType.APPLICATION_JSON));
    //add JWT token to the headers
    headers.add("Authorization", "JWT " + this.OAuth2Token);
    // build the request
    HttpEntity<SearchDTO> request = new HttpEntity<>(searchDTO, headers);

    // make an HTTP GET request with headers
    ResponseEntity<String> response = client.exchange(url, HttpMethod.POST, request, String.class);
    List<JournalTitleDTO> journalTitlesList = new ArrayList<>();
    // check response
    if (response.getStatusCode() == HttpStatus.OK) {
      final ObjectMapper mapper = new ObjectMapper();
      try {
        JsonNode neoJsonNode = mapper.readTree(response.getBody());
        JsonNode contentNodes = neoJsonNode.get("content");
        if (contentNodes instanceof ArrayNode) {
          for (JsonNode element : contentNodes) {
            JournalTitleDTO journal = new JournalTitleDTO();
            journal.setIssn(this.getTextValue(element, "issn"));
            journal.setMainTitle(this.getTextValue(element, "title"));
            journal.setPublisher(this.getTextValue(element, "publisher"));
            journal.setUrl(this.getTextValue(element, "url"));
            journal.setStartingYear(this.getTextValue(element, "date"));
            journal.setCountryCode(this.getTextValue(element, "country"));

            journalTitlesList.add(journal);
          }
        }
        int totalPages = neoJsonNode.get("totalPages").asInt();
        long totalElements = neoJsonNode.get("totalElements").asLong();
        CollectionPage page = new CollectionPage(searchDTO.getPage(), searchDTO.getSize(), totalPages, totalElements);
        return new Collection<>(journalTitlesList, page);
      } catch (JsonProcessingException e) {
        log.error("Error can't parse json response for the Title {}", searchDTO, e);
      }
    }
    CollectionPage page = new CollectionPage(searchDTO.getPage(), searchDTO.getSize(), 0, 0);
    return new Collection<>(journalTitlesList, page);
  }

  @SuppressWarnings("java:S2259")
  // transformKeyTitleFromISSN() cannot throw an NPE because the value is checked by the method StringTool.isNullOrEmpty()
  public Collection<JournalTitleDTO> getJournalTitlesByISSN(String issn) {
    String url = this.remoteServiceUrl + SolidifyConstants.URL_SEP + NOTICE + SolidifyConstants.URL_SEP + issn + "?natifjson=true";
    if (StringTool.isNullOrEmpty(this.OAuth2Token)) {
      this.getOAuth2TokenFromExternalService();
    }

    RestTemplateBuilder restTemplateBuilder = new RestTemplateBuilder();
    RestTemplate client = restTemplateBuilder.rootUri(url).build();
    // create headers
    HttpHeaders headers = new HttpHeaders();
    headers.setContentType(MediaType.APPLICATION_JSON);
    headers.setAccept(Collections.singletonList(MediaType.APPLICATION_JSON));
    //add JWT token to the headers
    headers.add("Authorization", "JWT " + this.OAuth2Token);
    // build the request
    HttpEntity<String> request = new HttpEntity<>(headers);
    List<JournalTitleDTO> journalTitlesList = new ArrayList<>();

    try {
      // make an HTTP GET request with headers
      ResponseEntity<String> response = client.exchange(url, HttpMethod.GET, request, String.class);

      JournalTitleDTO journalTitle = new JournalTitleDTO();
      journalTitle.setIssn(issn);
      // check response

      if (response.getStatusCode() == HttpStatus.OK) {
        final ObjectMapper mapper = new ObjectMapper();
        try {
          JsonNode neoJsonNode = mapper.readTree(response.getBody());
          JsonNode graphNodes = neoJsonNode.get("fields");
          if (graphNodes instanceof ArrayNode) {
            for (JsonNode element : graphNodes) {
              String properTitle = this.checkValueFromFieldAndSubfield(element, "245", "a");
              String pProperTitle = this.checkValueFromFieldAndSubfield(element, "245", "p");
              String nProperTitle = this.checkValueFromFieldAndSubfield(element, "245", "n");
              if (!StringTool.isNullOrEmpty(properTitle)) {
                //remove last point of the proper title
                if (properTitle.charAt(properTitle.length() - 1) == '.') {
                  properTitle = properTitle.substring(0, properTitle.length() -1);
                }
                if (StringTool.isNullOrEmpty(pProperTitle)) {
                  //delete all special characters
                  journalTitle.setMainTitle(properTitle.replaceAll("\\p{C}", ""));

                } else {
                  String title = nProperTitle == null ?
                          properTitle.concat(". " + pProperTitle) :
                          properTitle.concat(". " + nProperTitle + ", " + pProperTitle);
                  journalTitle.setMainTitle(title.replaceAll("\\p{C}", ""));
                }
              }
              String urlStr = this.checkValueFromFieldAndSubfield(element, "856", "u");
              if (!StringTool.isNullOrEmpty(urlStr)) {
                journalTitle.setUrl(urlStr);
              }

              String countryCode = this.checkValueFromFieldAndSubfield(element, "044", "c");
              if (!StringTool.isNullOrEmpty(countryCode)) {
                journalTitle.setCountryCode(countryCode);
              }

              String keyTitle = this.checkValueFromFieldAndSubfield(element, "222", "a");
              if (!StringTool.isNullOrEmpty(keyTitle)) {
                journalTitle.setKeytitle(keyTitle);
              }

              String publisher = this.checkValueFromFieldAndSubfield(element, "264", "b");
              if (!StringTool.isNullOrEmpty(publisher)) {
                journalTitle.setPublisher(publisher);
              }

              String issnl = this.checkValueFromFieldAndSubfield(element, "022", "l");
              if (!StringTool.isNullOrEmpty(issnl)) {
                journalTitle.setIssnl(issnl);
              }
            }
            journalTitlesList.add(journalTitle);
          }
        } catch (JsonProcessingException e) {
          log.error("Error can't parse json response for the ISSN {}", issn, e);
        }
      }

    } catch (
            RestClientException ex) { // added this to catch error from calling the external service
      log.error("Error with external research service for the ISSN {}", issn, ex);
    }

    // as the exepected result will be 1 or 0 at last, page and pageZie and totalPages are fixed to 1
    CollectionPage page = new CollectionPage(1, 1, 1, journalTitlesList.size());
    return new Collection<>(journalTitlesList, page);
  }

  private String checkValueFromFieldAndSubfield(JsonNode element, String field, String subfield) {
    if (element.get(field) != null && element.get(field).get("subfields") != null && element.get(field).get("subfields").isArray()) {
      for (JsonNode properKeyNode : element.get(field).get("subfields")) {
        if (properKeyNode.get(subfield) != null) {
          return properKeyNode.get(subfield).asText();
        }
      }
    }
    return null;
  }

  private String transformKeyTitleFromISSN(String title) {
    if (!StringTool.isNullOrEmpty(title)) {
      return title.replaceAll("(\\(Online\\)|\\(Print\\))", "");
    }
    return title;
  }

  private boolean nodeHasTextValue(JsonNode node) {
    return node != null && !node.isNull() && !StringTool.isNullOrEmpty(node.textValue());
  }

  private String cleanField(String field) {
    // Remove control characters U+0098 (START OF STRING), U+009C (STRING TERMINATOR) and trim
    if (field != null) {
      return field.replaceAll("\u0098", "")
              .replaceAll("\u009C", "")
              .trim();
    }
    return null;
  }

  private String getTextValue(JsonNode node, String fieldName) {
    if (this.nodeHasTextValue(node.get(fieldName))) {
      return this.cleanField(node.get(fieldName).asText());
    }
    return null;
  }

}
