package ch.unige.aou.task;

import java.time.OffsetDateTime;
import java.time.ZoneOffset;
import java.util.List;
import java.util.Objects;
import java.util.concurrent.ScheduledFuture;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.data.domain.PageRequest;
import org.springframework.data.domain.Pageable;
import org.springframework.data.domain.Sort;
import org.springframework.transaction.annotation.Transactional;

import ch.unige.solidify.exception.SolidifyRuntimeException;
import ch.unige.solidify.rest.ActionName;
import ch.unige.solidify.rest.CollectionPage;

import ch.unige.aou.business.NotificationService;
import ch.unige.aou.business.ScheduledTaskService;
import ch.unige.aou.model.notification.Notification;
import ch.unige.aou.model.notification.NotificationType;
import ch.unige.aou.model.schedule.ScheduledTask;
import ch.unige.aou.specification.NotificationSpecification;

public abstract class AbstractTaskRunnable implements Runnable {

  private static final Logger log = LoggerFactory.getLogger(AbstractTaskRunnable.class);

  protected ScheduledTaskService scheduledTaskService;

  protected ScheduledTask scheduledTask;
  protected ScheduledFuture<AbstractTaskRunnable> scheduledFuture;
  protected final NotificationService notificationService;

  protected final TaskExecutorService taskExecutorService;

  /**
   * Variable used to be able to use the previous task run start time.
   * Required if the task execution needs this value as its begins by storing a new value.
   */
  protected OffsetDateTime lastExecutionStartTime = null;

  protected AbstractTaskRunnable(ScheduledTaskService scheduledTaskService, ScheduledTask scheduledTask, NotificationService notificationService,
          TaskExecutorService taskExecutorService) {
    if (scheduledTask == null) {
      throw new SolidifyRuntimeException("ScheduledTask cannot be null");
    }

    this.scheduledTaskService = scheduledTaskService;
    this.scheduledTask = scheduledTask;
    this.notificationService = notificationService;
    this.taskExecutorService = taskExecutorService;

    this.copyLastExecutionStartTime();
  }

  public ScheduledTask getScheduledTask() {
    return this.scheduledTask;
  }

  public ScheduledFuture<AbstractTaskRunnable> getScheduledFuture() {
    return this.scheduledFuture;
  }

  public void setScheduledFuture(ScheduledFuture<AbstractTaskRunnable> scheduledFuture) {
    this.scheduledFuture = scheduledFuture;
  }

  @Override
  @Transactional
  public void run() {
    log.info("Task '{}' ({}) starting", this.scheduledTask.getResId(), this.scheduledTask.getName());
    this.copyLastExecutionStartTime();
    this.updateLastExecutionStartTime();
    this.executeTask();
    this.updateLastExecutionEndTime();
    log.info("Task '{}' ({}) is over", this.scheduledTask.getResId(), this.scheduledTask.getName());
  }

  protected void updateLastExecutionStartTime() {
    this.scheduledTask.setLastExecutionStart(OffsetDateTime.now());
    this.saveScheduledTask(this.scheduledTask);
  }

  protected void copyLastExecutionStartTime() {
    this.lastExecutionStartTime = this.scheduledTask.getLastExecutionStart();
  }

  protected void updateLastExecutionEndTime() {
    this.scheduledTask.setLastExecutionEnd(OffsetDateTime.now());
    this.saveScheduledTask(this.scheduledTask);
  }

  /**
   * Get notifications of the given type that have been created since the last execution of the task and have not been sent yet
   *
   * @param notificationType
   * @return
   */
  protected List<Notification> getNotificationsToSend(NotificationType notificationType) {
    if (this.lastExecutionStartTime == null) {
      this.lastExecutionStartTime = OffsetDateTime.of(2000, 1, 1, 0, 0, 0, 0, ZoneOffset.UTC);
    }

    OffsetDateTime searchFrom = this.lastExecutionStartTime.minusSeconds(10);

    Notification search = new Notification();
    search.setNotificationType(notificationType);
    NotificationSpecification notificationSpecification = this.notificationService.getSpecification(search);
    notificationSpecification.setCreatedDateMinValue(searchFrom);
    notificationSpecification.setOnlyWithNullSentTime(true);
    Pageable pageable = PageRequest.of(0, CollectionPage.MAX_SIZE_PAGE, Sort.by(ActionName.CREATION));
    return this.notificationService.findAll(notificationSpecification, pageable).toList();
  }

  /**
   * Update the scheduled task, but only if it hasn't been deleted while it was running
   *
   * @param scheduledTask
   */
  protected void saveScheduledTask(ScheduledTask scheduledTask) {
    if (this.scheduledTaskService.existsById(scheduledTask.getResId())) {
      this.scheduledTaskService.save(scheduledTask);
    } else {
      log.warn("The scheduled task '{}' does not exist anymore", scheduledTask.getResId());
    }
  }

  protected abstract void executeTask();

  @Override
  public boolean equals(Object o) {
    if (this == o) {
      return true;
    }
    if (o == null || this.getClass() != o.getClass()) {
      return false;
    }

    AbstractTaskRunnable that = (AbstractTaskRunnable) o;
    if (this.scheduledTask != null) {
      return Objects.equals(this.scheduledTask.getResId(), that.scheduledTask.getResId());
    }

    return false;
  }

  @Override
  public int hashCode() {
    return 31 * (this.scheduledTask != null ? this.scheduledTask.hashCode() : 0);
  }
}
