package ch.unige.aou.task;

import java.time.LocalDateTime;
import java.time.OffsetDateTime;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import javax.transaction.Transactional;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.data.domain.PageRequest;
import org.springframework.data.domain.Pageable;
import org.springframework.data.domain.Sort;
import org.springframework.stereotype.Service;

import ch.unige.solidify.config.SolidifyEventPublisher;
import ch.unige.solidify.rest.ActionName;
import ch.unige.solidify.rest.CollectionPage;
import ch.unige.solidify.rest.ResourceNormalized;
import ch.unige.solidify.service.MessageService;

import ch.unige.aou.business.NotificationService;
import ch.unige.aou.business.PersonService;
import ch.unige.aou.business.PublicationService;
import ch.unige.aou.business.ResearchGroupService;
import ch.unige.aou.business.UserService;
import ch.unige.aou.config.AouProperties;
import ch.unige.aou.message.EmailMessage;
import ch.unige.aou.model.notification.Notification;
import ch.unige.aou.model.notification.NotificationType;
import ch.unige.aou.model.publication.Publication;
import ch.unige.aou.model.rest.ResearchGroupDTO;
import ch.unige.aou.model.security.User;
import ch.unige.aou.model.settings.Person;
import ch.unige.aou.service.AouService;

@Service
public class TaskExecutorService extends AouService {

  private static final Logger log = LoggerFactory.getLogger(TaskExecutorService.class);

  protected final NotificationService notificationService;
  private final PublicationService publicationService;
  private final PersonService personService;
  private final UserService userService;
  private final ResearchGroupService researchGroupService;
  private final int delaySinceLastCheck;

  public TaskExecutorService(MessageService messageService, PublicationService publicationService, NotificationService notificationService,
          AouProperties aouProperties, PersonService personService, UserService userService, ResearchGroupService researchGroupService) {
    super(messageService);
    this.publicationService = publicationService;
    this.delaySinceLastCheck = aouProperties.getDelaySinceLastCheck();
    this.notificationService = notificationService;
    this.personService = personService;
    this.userService = userService;
    this.researchGroupService = researchGroupService;
  }

  /**
   * Group notifications by person and keep them only if the related publication has still the given status. (we don't send a notification if
   * the publication status has changed since the notification was created as it wouldn't make sense anymore)
   *
   * @param notificationsToSend
   * @param filterStatus
   * @return
   */
  protected Map<String, List<Notification>> getFilteredNotificationsByPersonId(List<Notification> notificationsToSend,
          List<Publication.PublicationStatus> filterStatuses) {

    Map<String, List<Notification>> notificationsByPersonId = new HashMap<>();
    for (Notification notificationToSend : notificationsToSend) {

      // keep the notification only if it has the specified filter status
      if (filterStatuses.stream().anyMatch(status -> status == notificationToSend.getEvent().getPublication().getStatus())) {
        String recipientId = notificationToSend.getRecipient().getResId();
        notificationsByPersonId.computeIfAbsent(recipientId, s -> new ArrayList<>());
        notificationsByPersonId.get(recipientId).add(notificationToSend);
      }
    }
    return notificationsByPersonId;
  }

  /**
   * Create and send an email to the person if he has subscribed to the notification type, with publication ids linked to the notifications.
   * The sending process uses a JMS queue.
   *
   * @param personId
   * @param notifications
   * @param notificationType
   * @param emailTemplate
   */
  private void sendEmailWithPublicationIdsIfUserHasSubscribedToNotificationType(String personId, List<Notification> notifications,
          NotificationType notificationType, EmailMessage.EmailTemplate emailTemplate) {

    if (this.personService.hasSubscribedToNotificationType(personId, notificationType)) {

      // validator's email
      User recipientUser = this.userService.findByPersonResId(personId);

      // compute a list of unique publication ids to notify (to not include the same publication more than once in an email)
      String[] uniquePublicationIds = notifications.stream().map(notification -> notification.getEvent().getPublication().getResId()).distinct()
              .toArray(String[]::new);

      // send email
      SolidifyEventPublisher.getPublisher().publishEvent(new EmailMessage(recipientUser.getEmail(), emailTemplate, uniquePublicationIds));

      // mark all notifications as sent
      OffsetDateTime sentTime = OffsetDateTime.now();
      for (Notification notification : notifications) {
        this.notificationService.updateSentTime(notification.getResId(), sentTime);
        log.info("notification {} sentTime updated with value {}", notification.getResId(), sentTime);
      }
    }
  }

  private void sendEmailToValidatorForForgottenPublications(String personId, List<Publication> value,
          EmailMessage.EmailTemplate emailTemplate) {

    // validator's email
    User recipientUser = this.userService.findByPersonResId(personId);

    // compute a list of unique publication ids to notify (to not include the same publication more than once in an email)
    String[] uniquePublicationIds = value.stream().map(ResourceNormalized::getResId).distinct().toArray(String[]::new);

    // send email
    SolidifyEventPublisher.getPublisher().publishEvent(new EmailMessage(recipientUser.getEmail(), emailTemplate, uniquePublicationIds));
  }

  /********************************************************************************************************************************/

  @Transactional
  public void notifyForgottenPublicationsToValidators() {
    OffsetDateTime offsetDelay = OffsetDateTime.now().minusDays(this.delaySinceLastCheck);
    Pageable pageable = PageRequest.of(0, CollectionPage.MAX_SIZE_PAGE, Sort.by(ActionName.CREATION));
    List<Publication> publications = this.publicationService.findPublicationByStatusAndLastUpdate(Publication.PublicationStatus.IN_VALIDATION,
            offsetDelay, pageable).toList();

    Map<Person, List<Publication>> publicationsByValidator = new HashMap<>();
    for (Publication publication : publications) {
      final Publication managedPublication = this.publicationService.findOne(publication.getResId());
      List<Person> validators = this.personService.getValidators(publication);
      validators.stream().forEach(validator -> {
        publicationsByValidator.computeIfAbsent(validator, k -> new ArrayList<>()).add(managedPublication);
      });
    }

    for (Map.Entry<Person, List<Publication>> entry : publicationsByValidator.entrySet()) {
      this.sendEmailToValidatorForForgottenPublications(entry.getKey().getResId(), entry.getValue(),
              EmailMessage.EmailTemplate.FORGOTTEN_PUBLICATIONS_TO_VALIDATE);
    }
  }

  @Transactional
  public void notifyPublicationsRejectedToDepositor(List<Notification> notificationsToSend) {
    Map<String, List<Notification>> notificationsByPersonId = this
            .getFilteredNotificationsByPersonId(notificationsToSend, List.of(Publication.PublicationStatus.REJECTED));

    for (Map.Entry<String, List<Notification>> entry : notificationsByPersonId.entrySet()) {
      this.sendEmailWithPublicationIdsIfUserHasSubscribedToNotificationType(entry.getKey(), entry.getValue(),
              NotificationType.MY_PUBLICATION_REJECTED, EmailMessage.EmailTemplate.MY_PUBLICATIONS_REJECTED);
    }
  }

  @Transactional
  public void notifyPublicationsToValidate(List<Notification> notificationsToSend) {
    Map<String, List<Notification>> notificationsByPersonId = this
            .getFilteredNotificationsByPersonId(notificationsToSend, List.of(Publication.PublicationStatus.IN_VALIDATION));

    for (Map.Entry<String, List<Notification>> entry : notificationsByPersonId.entrySet()) {
      this.sendEmailWithPublicationIdsIfUserHasSubscribedToNotificationType(entry.getKey(), entry.getValue(),
              NotificationType.PUBLICATION_TO_VALIDATE, EmailMessage.EmailTemplate.NEW_DEPOSITS_TO_VALIDATE);
    }
  }

  @Transactional
  public void notifyPublicationsValidatedToContributors(List<Notification> notificationsToSend) {
    Map<String, List<Notification>> notificationsByPersonId = this
            .getFilteredNotificationsByPersonId(notificationsToSend, List.of(Publication.PublicationStatus.COMPLETED));

    for (Map.Entry<String, List<Notification>> entry : notificationsByPersonId.entrySet()) {
      this.sendEmailWithPublicationIdsIfUserHasSubscribedToNotificationType(entry.getKey(), entry.getValue(),
              NotificationType.MY_INDIRECT_PUBLICATION_VALIDATED, EmailMessage.EmailTemplate.MY_INDIRECT_PUBLICATION_VALIDATED);
    }
  }

  @Transactional
  public void notifyPublicationsValidatedToDepositor(List<Notification> notificationsToSend) {
    Map<String, List<Notification>> notificationsByPersonId = this
            .getFilteredNotificationsByPersonId(notificationsToSend, List.of(Publication.PublicationStatus.COMPLETED));

    for (Map.Entry<String, List<Notification>> entry : notificationsByPersonId.entrySet()) {
      this.sendEmailWithPublicationIdsIfUserHasSubscribedToNotificationType(entry.getKey(), entry.getValue(),
              NotificationType.MY_PUBLICATION_VALIDATED, EmailMessage.EmailTemplate.MY_PUBLICATIONS_VALIDATED);
    }
  }

  @Transactional
  public void notifyValidatorsOnComments(List<Notification> notificationsToSend) {
    Map<String, List<Notification>> notificationsByPersonId = this.getFilteredNotificationsByPersonId(notificationsToSend,
            List.of(Publication.PublicationStatus.IN_VALIDATION, Publication.PublicationStatus.FEEDBACK_REQUIRED));

    for (Map.Entry<String, List<Notification>> entry : notificationsByPersonId.entrySet()) {
      this.sendEmailWithPublicationIdsIfUserHasSubscribedToNotificationType(entry.getKey(), entry.getValue(),
              NotificationType.PUBLICATION_TO_VALIDATE_COMMENTED, EmailMessage.EmailTemplate.NEW_COMMENT_FOR_VALIDATORS_DEPOSIT);
    }
  }

  @Transactional
  public void synchronizeFromUnigeReferential(LocalDateTime createdOrModifiedSince) {
    List<ResearchGroupDTO> synchronizedResearchGroups = this.researchGroupService.synchronizeFromUnigeReferential(createdOrModifiedSince);

    if (!synchronizedResearchGroups.isEmpty()) {
      List<User> adminOrRootUsers = this.userService.findAdminOrRootUsers();

      // build research groups strings
      List<String> researchGroupsIds = new ArrayList<>();
      for (ResearchGroupDTO researchGroupDTO : synchronizedResearchGroups) {
        researchGroupsIds.add(researchGroupDTO.getId());
      }
      String[] researchGroupsIdsArray = researchGroupsIds.stream().toArray(String[]::new);

      // send emails
      for (User user : adminOrRootUsers) {
        SolidifyEventPublisher.getPublisher().publishEvent(
                new EmailMessage(user.getEmail(), EmailMessage.EmailTemplate.SYNCHRONIZED_RESEARCH_GROUPS, researchGroupsIdsArray));
      }
    }
  }
}
