package ch.unige.aou.test;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertFalse;
import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.junit.jupiter.api.Assertions.fail;

import java.io.IOException;
import java.io.InputStream;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.xpath.XPathExpressionException;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.extension.ExtendWith;
import org.mockito.Mock;
import org.springframework.context.annotation.Profile;
import org.springframework.core.io.ClassPathResource;
import org.springframework.test.context.junit.jupiter.SpringExtension;
import org.w3c.dom.Document;
import org.xml.sax.SAXException;

import ch.unige.solidify.config.SolidifyProperties;
import ch.unige.solidify.exception.SolidifyRuntimeException;
import ch.unige.solidify.service.MessageService;
import ch.unige.solidify.util.FileTool;
import ch.unige.solidify.util.XMLTool;

import ch.unige.aou.AouConstants;
import ch.unige.aou.business.ContributorService;
import ch.unige.aou.business.DocumentFileService;
import ch.unige.aou.business.DocumentFileTypeService;
import ch.unige.aou.config.AouProperties;
import ch.unige.aou.model.xml.deposit.v2_3.Contributor;
import ch.unige.aou.model.xml.deposit.v2_3.DateTypes;
import ch.unige.aou.model.xml.deposit.v2_3.DepositDoc;
import ch.unige.aou.service.MetadataService;
import ch.unige.aou.service.metadata.imports.DoiCrossrefImportService;
import ch.unige.aou.service.rest.JournalTitleRemoteService;
import ch.unige.aou.service.rest.OpenAireService;

@Profile({ "test", "sec-noauth" })
@ExtendWith(SpringExtension.class)
class DoiCrossrefImportServiceTest {

  private static final String CROSSREFF_XML_METADATA_FILE = "test-crossref-response.xml";
  private static final String TEST_DOI = "10.1126/science.1175088";

  private DoiCrossrefImportService doiCrossrefImportService;

  @Mock
  private MessageService messageService;

  @Mock
  protected MetadataService metadataService;

  @Mock
  protected ContributorService contributorService;

  @Mock
  protected DocumentFileService documentFileService;

  @Mock
  protected DocumentFileTypeService documentFileTypeService;

  @Mock
  protected JournalTitleRemoteService journalTitleRemoteService;

  @Mock
  protected OpenAireService openAireService;

  @BeforeEach
  public void setUp() {
    AouProperties aouProperties = new AouProperties(new SolidifyProperties());
    this.doiCrossrefImportService = new DoiCrossrefImportService(aouProperties, this.messageService, this.metadataService,
            this.contributorService, this.documentFileService, this.documentFileTypeService, this.journalTitleRemoteService,
            this.openAireService);
  }

  @Test
  void fillDepositDocTest() {

    try {
      Document crossrefDoc = XMLTool.parseXML(this.getCrossrefXmlMetadata());
      DepositDoc depositDoc = new DepositDoc();
      this.doiCrossrefImportService.fillDepositDoc(TEST_DOI, depositDoc, crossrefDoc);

      assertEquals("An ER-Mitochondria Tethering Complex Revealed by a Synthetic Biology Screen", depositDoc.getTitle().getContent());
      assertEquals(AouConstants.LANG_CODE_ENGLISH, depositDoc.getTitle().getLang());
      assertEquals(AouConstants.LANG_CODE_ENGLISH, depositDoc.getLanguages().getLanguage().get(0));
      assertNotNull(depositDoc.getDates());
      assertNotNull(depositDoc.getDates().getDate());
      assertFalse(depositDoc.getDates().getDate().isEmpty());
      assertEquals("2009-07-24", depositDoc.getDates().getDate().get(0).getContent());
      assertEquals(DateTypes.PUBLICATION, depositDoc.getDates().getDate().get(0).getType());
      assertEquals("2009-06-25", depositDoc.getDates().getDate().get(1).getContent());
      assertEquals(DateTypes.FIRST_ONLINE, depositDoc.getDates().getDate().get(1).getType());
      assertNotNull(depositDoc.getContributors());
      assertNotNull(depositDoc.getContributors().getContributorOrCollaboration());
      assertEquals(7, depositDoc.getContributors().getContributorOrCollaboration().size());
      assertTrue(depositDoc.getContributors().getContributorOrCollaboration().get(0) instanceof Contributor);
      assertEquals("Kornmann", ((Contributor) depositDoc.getContributors().getContributorOrCollaboration().get(0)).getLastname());
      assertEquals("Benoît", ((Contributor) depositDoc.getContributors().getContributorOrCollaboration().get(0)).getFirstname());
      assertTrue(depositDoc.getPages().getPagingOrOther().get(0).getName().toString().contains("paging"));
      assertEquals("477-481", depositDoc.getPages().getPagingOrOther().get(0).getValue());
      assertEquals(TEST_DOI, depositDoc.getIdentifiers().getDoi());
      assertEquals("0036-8075", depositDoc.getIdentifiers().getIssn());
      assertEquals("https://www.sciencemag.org/lookup/doi/10.1126/science.1175088", depositDoc.getPublisherVersionUrl());
      assertEquals("Science", depositDoc.getContainer().getTitle());
      assertEquals("325", depositDoc.getContainer().getVolume().toString());
      assertEquals("5939", depositDoc.getContainer().getIssue().toString());

    } catch (ParserConfigurationException | IOException | SAXException | XPathExpressionException e) {
      throw new SolidifyRuntimeException("unable to search DOI by Crossref", e);
    }
  }

  private String getCrossrefXmlMetadata() {
    try (InputStream jsonStream = new ClassPathResource(CROSSREFF_XML_METADATA_FILE).getInputStream()) {
      return FileTool.toString(jsonStream);
    } catch (IOException e) {
      fail("Unable to read " + CROSSREFF_XML_METADATA_FILE + ": " + e.getMessage());
      return null;
    }
  }
}
