package ch.unige.aou.service;

import java.util.Collection;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.GrantedAuthority;
import org.springframework.security.core.context.SecurityContextHolder;

import ch.unige.solidify.auth.model.ApplicationRole;
import ch.unige.solidify.auth.model.AuthApplicationRole;
import ch.unige.solidify.exception.SolidifyRuntimeException;
import ch.unige.solidify.util.StringTool;

import ch.unige.aou.business.PersonService;
import ch.unige.aou.controller.AouControllerAction;

public abstract class AbstractPermissionService {

  private static final Logger logger = LoggerFactory.getLogger(AbstractPermissionService.class);

  protected PersonService personService;

  protected AbstractPermissionService(PersonService personService) {
    this.personService = personService;
  }

  /******************************************************************************************************/

  public boolean isUserConnectedLinkedToPerson(String personId) {
    if (StringTool.isNullOrEmpty(personId)) {
      return false;
    }
    return personId.equals(this.getPersonAuthenticatedId());
  }

  /******************************************************************************************************/

  protected boolean canBypassPermissionCheck() {
    final String[] authorizedApplicationRoleNames = { AuthApplicationRole.ADMIN_ID, AuthApplicationRole.ROOT_ID,
            AuthApplicationRole.TRUSTED_CLIENT_ID };

    final Collection<? extends GrantedAuthority> authorities = this.getAuthentication().getAuthorities();

    for (final String roleName : authorizedApplicationRoleNames) {
      for (final GrantedAuthority authority : authorities) {
        if (authority.toString().equals(roleName)) {
          return true;
        }
      }
    }
    return false;
  }

  protected boolean hasUserApplicationRole() {
    return this.hasApplicationRole(AuthApplicationRole.USER);
  }

  protected boolean hasApplicationRole(ApplicationRole applicationRole) {
    final Collection<? extends GrantedAuthority> authorities = this.getAuthentication().getAuthorities();
    return authorities.stream().anyMatch(a -> a.toString().equals(applicationRole.getResId()));
  }

  protected Authentication getAuthentication() {
    return SecurityContextHolder.getContext().getAuthentication();
  }

  protected String getPersonAuthenticatedId() {
    return this.personService.getLinkedPersonId(this.getAuthentication());
  }

  protected AouControllerAction getControllerAction(Object permission) {
    try {
      return AouControllerAction.valueOf(permission.toString());
    } catch (final Exception ex) {
      logger.error("Unable to check for access permission: the action name '{}' is wrong", permission);
      throw new SolidifyRuntimeException("unable to check for access permission: the action name '" + permission + "' is wrong");
    }
  }

  protected Object getResourceToCheckPermissionOn(Object existingResource, Object updateResource) {
    if (existingResource != null) {
      // Case of UPDATE
      return existingResource;
    } else if (updateResource != null) {
      // Case of CREATE
      return updateResource;
    } else {
      throw new SolidifyRuntimeException("error while checking user permissions on resource: no given resource to check");
    }
  }
}
